/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "analyze.h"

/**
 * @description: 分析器，进行语义分析和查询重写，需要检查不符合语义规定的部分
 * @param {shared_ptr<ast::TreeNode>} parse parser生成的结果集
 * @return {shared_ptr<Query>} Query 
 */
std::shared_ptr<Query> Analyze::do_analyze(std::shared_ptr<ast::TreeNode> parse)
{
    std::shared_ptr<Query> query = std::make_shared<Query>();
    query->having=std::make_shared<Having>();
    if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(parse))
    {
        // 处理表名
        query->tables = std::move(x->tabs);
        /** TODO: 检查表是否存在 */

        for(int i=0;i<(query->tables).size();i++){
            std::string table_name=(query->tables)[i];
            if(!sm_manager_->db_.is_table(table_name)){
                throw TableNotFoundError(table_name);
            }
        }

        // 处理target list，再target list中添加上表名，例如 a.id
        for (auto &sv_sel_col : x->cols) {
            TabCol sel_col = {.tab_name = sv_sel_col->tab_name, .col_name = sv_sel_col->col_name};
            query->cols.push_back(sel_col);
        }
        // 这里目前也有问题，不能把表名加上去
        std::vector<ColMeta> all_cols;
        get_all_cols(query->tables, all_cols);
        if (query->cols.empty()) {
            // select all columns
            for (auto &col : all_cols) {
                TabCol sel_col = {.tab_name = col.tab_name, .col_name = col.name};
                query->cols.push_back(sel_col);
            }
        } else {
            // infer table name from column name
            for (auto &sel_col : query->cols) {
                sel_col = check_column(all_cols, sel_col);  // 列元数据校验
            }
        }
        //处理where条件
        get_clause(x->conds, query->conds);
        check_clause(query->tables, query->conds);
        //含有聚合的健壮性测试
        stronger_check(x);
        //处理having
        if(!x->havingconds.empty())
        {
            get_havingconds(x->havingconds,query->having);
            check_clause(query->tables,query->having->conds);
            check_havingconds(query->tables,query->having->havingconds);
        }
    } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(parse)) {
        /** TODO: */
        /** TODO: 检查表是否存在 */
        if(!sm_manager_->db_.is_table(x->tab_name)){
            throw TableNotFoundError(x->tab_name);
        }
        //处理set
        for(auto it=x->set_clauses.begin();it!=x->set_clauses.end();it++){
            if(it->get()->is_add == true){
                query->set_clauses.push_back({{x->tab_name,it->get()->col_name},convert_sv_value(it->get()->val), true});
            }
            else{
                query->set_clauses.push_back({{x->tab_name,it->get()->col_name},convert_sv_value(it->get()->val), false});   
            }
        }
        //处理where条件
        get_clause(x->conds,query->conds);
        //向query中插入表名
        check_clause({x->tab_name},query->conds);
    } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(parse)) {
        //处理where条件
        get_clause(x->conds, query->conds);
        check_clause({x->tab_name}, query->conds);        
    } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(parse)) {
        // 处理insert 的values值
        for (auto &sv_val : x->vals) {
            query->values.push_back(convert_sv_value(sv_val));
        }
    } else {
        // do nothing
    }
    query->parse = std::move(parse);
    
    return query;
}

void Analyze::stronger_check(const std::shared_ptr<ast::SelectStmt> parse_tree){
    //1.select列表中不能出现不再groupby子句中的非聚集列
    bool is_ok=true;
    if(!parse_tree->cols.empty()){
        if(parse_tree->aggregatestmts.size()==0&&(parse_tree->groupby!=nullptr||parse_tree->havingconds.size()!=0)){
            is_ok=false;
            throw AggError("invalid col");
        }
        if(parse_tree->groupby!=nullptr){
            for(auto &agg_col:parse_tree->agg_cols){
                if(auto col=std::dynamic_pointer_cast<ast::Col>(agg_col)){
                    int i;
                    for(i=0;i<parse_tree->groupby->cols.size();i++){
                        if(col->col_name==parse_tree->groupby->cols[i]->col_name&&
                            col->tab_name==parse_tree->groupby->cols[i]->tab_name){
                                break;
                            }
                    }
                    if(i==parse_tree->groupby->cols.size()){
                        is_ok=false;
                        throw AggError("col not in groupby");
                    }
                }
            }
        }
        if(parse_tree->havingconds.size()!=0){
            for(auto &havingcond:parse_tree->havingconds){
                if(auto col=std::dynamic_pointer_cast<ast::Col>(havingcond->lhs)){
                    int i=0;
                    for(i=0;i<parse_tree->groupby->cols.size();i++){
                    if(col->col_name==parse_tree->groupby->cols[i]->col_name&&
                        col->tab_name==parse_tree->groupby->cols[i]->tab_name){
                            break;
                        }
                    }
                    if(i==parse_tree->groupby->cols.size()){
                        throw AggError("having col not in group by");
                        is_ok=false;
                    }
                }
            }
            
        }
        if(is_ok==false){
            std::fstream outfile;
            outfile.open("output.txt",std::ios::out | std::ios::app);
            outfile << "failure\n";
            outfile.close();
        }
    }
}

TabCol Analyze::check_column(const std::vector<ColMeta> &all_cols, TabCol target) {
    if (target.tab_name.empty()) {
        // Table name not specified, infer table name from column name
        std::string tab_name;
        for (auto &col : all_cols) {
            if (col.name == target.col_name) {
                if (!tab_name.empty()) {
                    throw AmbiguousColumnError(target.col_name);
                }
                tab_name = col.tab_name;
            }
        }
        if (tab_name.empty()) {
            throw ColumnNotFoundError(target.col_name);
        }
        target.tab_name = tab_name;
    } else {
        /** TODO: Make sure target column exists */
        bool exist = false;
        for (auto &col : all_cols) 
        {   
            //select stu.id from stu,stu1,stu2...
            //查找被查找表stu,stu1...中所有列是否有查找列id和对应表stu
            if (col.name == target.col_name && col.tab_name == target.tab_name) 
            {   
                exist = true;
                break;
            }
            else
            {
                continue;
            }
        }
        if(!exist)
        {   
            //没找到说明可能是select a.id from b,就算b表有id列，但是表名不匹配也说明错误，抛出未找到列的异常
            throw ColumnNotFoundError(target.col_name);
        }
    }
    return target;
}

void Analyze::get_all_cols(const std::vector<std::string> &tab_names, std::vector<ColMeta> &all_cols) {
    for (auto &sel_tab_name : tab_names) {
        // 这里db_不能写成get_db(), 注意要传指针
        const auto &sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
        all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
    }
}

void Analyze::get_havingconds(const std::vector<std::shared_ptr<ast::HavingExpr>> havingexprs,std::shared_ptr<Having>& having){
    if(!having->conds.empty()){
        having->conds.clear();
    }
    if(!having->havingconds.empty()){
        having->havingconds.clear();
    }
    
    for(auto &havingexpr:havingexprs){
        if(auto col=std::dynamic_pointer_cast<ast::Col>(havingexpr->lhs)){
            Condition cond;
            cond.lhs_col = {.tab_name = col->tab_name, .col_name = col->col_name};
            cond.op = convert_sv_comp_op(havingexpr->op);
            if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(havingexpr->rhs)) {
                cond.is_rhs_val = true;
                cond.rhs_val = convert_sv_value(rhs_val);
            } else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(havingexpr->rhs)) {
                cond.is_rhs_val = false;
                cond.rhs_col = {.tab_name = rhs_col->tab_name, .col_name = rhs_col->col_name};
            }
            having->conds.push_back(cond);
        }else if(auto aggregate=std::dynamic_pointer_cast<ast::AggregateStmt>(havingexpr->lhs)){
            TabCol aggfunction_col;
            if(aggregate->col==nullptr){
                aggregate->col=std::make_shared<ast::Col>("","*");
            }
            aggfunction_col = {.tab_name = aggregate->col->tab_name, .col_name = aggregate->col->col_name};
            AggregateFunction aggfunc(aggregate->type,aggregate->as_col,aggfunction_col);
            bool is_rhs_val;
            Value cond_rhs_val;
            // cond.op = convert_sv_comp_op(havingexpr->op);
            if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(havingexpr->rhs)) {
                is_rhs_val = true;
                cond_rhs_val = convert_sv_value(rhs_val);
            }
            TabCol rhs_col;
            HavingCondition havingcond(aggfunc,convert_sv_comp_op(havingexpr->op),is_rhs_val,rhs_col,cond_rhs_val);
            having->havingconds.push_back(havingcond);
        }
    }
}

void Analyze::get_clause(const std::vector<std::shared_ptr<ast::BinaryExpr>> &sv_conds, std::vector<Condition> &conds) {
    conds.clear();
    for (auto &expr : sv_conds) {
        Condition cond;
        cond.lhs_col = {.tab_name = expr->lhs->tab_name, .col_name = expr->lhs->col_name};
        cond.op = convert_sv_comp_op(expr->op);
        if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs)) {
            cond.is_rhs_val = true;
            cond.rhs_val = convert_sv_value(rhs_val);
        } else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(expr->rhs)) {
            cond.is_rhs_val = false;
            cond.rhs_col = {.tab_name = rhs_col->tab_name, .col_name = rhs_col->col_name};
        }
        conds.push_back(cond);
    }
}

void Analyze::check_clause(const std::vector<std::string> &tab_names, std::vector<Condition> &conds) {
    // auto all_cols = get_all_cols(tab_names);
    std::vector<ColMeta> all_cols;
    get_all_cols(tab_names, all_cols);
    // Get raw values in where clause
    for (auto &cond : conds) {
        // Infer table name from column name
        cond.lhs_col = check_column(all_cols, cond.lhs_col);
        if (!cond.is_rhs_val) {
            cond.rhs_col = check_column(all_cols, cond.rhs_col);
        }
        TabMeta &lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
        auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
        ColType lhs_type = lhs_col->type;
        ColType rhs_type;
        if (cond.is_rhs_val) {
            
            //可能存在类型转化
            if(cond.rhs_val.type==TYPE_INT&&lhs_col->type==TYPE_FLOAT){
                float temp=(float)cond.rhs_val.int_val;
                cond.rhs_val.set_float(temp);
                // cond.rhs_val.init_raw(lhs_col->len);
            }
            cond.rhs_val.init_raw(lhs_col->len);
            rhs_type = cond.rhs_val.type;
        } else {
            TabMeta &rhs_tab = sm_manager_->db_.get_table(cond.rhs_col.tab_name);
            auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
            rhs_type = rhs_col->type;
        }
        if (lhs_type != rhs_type) {
            throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
        }
    }
}

void Analyze::check_havingconds(const std::vector<std::string> &tab_names, std::vector<HavingCondition>& havingconds) {
    std::vector<ColMeta> all_cols;
    get_all_cols(tab_names, all_cols);
    // Get raw values in where clause
    for (auto &havingcond : havingconds) {
        if(havingcond.lhs_col.agg_col.col_name=="*"){
            ColType lhs_type = TYPE_INT;
            ColType rhs_type;
            if (havingcond.is_rhs_val) {
                havingcond.rhs_val.init_raw(sizeof(int));
                rhs_type = havingcond.rhs_val.type;
            } 
            if (lhs_type != rhs_type) {
                throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
            }
        }
        else{
            havingcond.lhs_col.agg_col = check_column(all_cols, havingcond.lhs_col.agg_col);
            if (!havingcond.is_rhs_val) {
                // cond.rhs_col = check_column(all_cols, cond.rhs_col);
            }
            TabMeta &lhs_tab = sm_manager_->db_.get_table(havingcond.lhs_col.agg_col.tab_name);
            auto lhs_col = lhs_tab.get_col(havingcond.lhs_col.agg_col.col_name);
            ColType lhs_type = lhs_col->type;
            ColType rhs_type;
            if (havingcond.is_rhs_val) {
                //可能存在类型转化
                if(havingcond.rhs_val.type==TYPE_INT&&lhs_col->type==TYPE_FLOAT){
                    float temp=(float)havingcond.rhs_val.int_val;
                    havingcond.rhs_val.set_float(temp);
                    // cond.rhs_val.init_raw(lhs_col->len);
                }
                havingcond.rhs_val.init_raw(lhs_col->len);
                rhs_type = havingcond.rhs_val.type;
            } 
            if (lhs_type != rhs_type) {
                throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
            }
        }
    }
}

Value Analyze::convert_sv_value(const std::shared_ptr<ast::Value> &sv_val) {
    Value val;
    if (auto int_lit = std::dynamic_pointer_cast<ast::IntLit>(sv_val)) {
        val.set_int(int_lit->val);
    } else if (auto float_lit = std::dynamic_pointer_cast<ast::FloatLit>(sv_val)) {
        val.set_float(float_lit->val);
    } else if (auto str_lit = std::dynamic_pointer_cast<ast::StringLit>(sv_val)) {
        val.set_str(str_lit->val);
    } else {
        throw InternalError("Unexpected sv value type");
    }
    return val;
}

CompOp Analyze::convert_sv_comp_op(ast::SvCompOp op) {
    std::map<ast::SvCompOp, CompOp> m = {
        {ast::SV_OP_EQ, OP_EQ}, {ast::SV_OP_NE, OP_NE}, {ast::SV_OP_LT, OP_LT},
        {ast::SV_OP_GT, OP_GT}, {ast::SV_OP_LE, OP_LE}, {ast::SV_OP_GE, OP_GE},
    };
    return m.at(op);
}
