/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class DeleteExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;                   // 表的元数据
    std::vector<Condition> conds_;  // delete的条件
    RmFileHandle *fh_;              // 表的数据文件句柄
    std::vector<Rid> rids_;         // 需要删除的记录的位置
    std::string tab_name_;          // 表名称
    SmManager *sm_manager_;

   public:
    DeleteExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<Condition> conds,
                   std::vector<Rid> rids, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        context_ = context;
    }

    std::unique_ptr<RmRecord> Next() override {
        if(context_!=nullptr){
            context_->lock_mgr_->lock_exclusive_on_table(context_->txn_, fh_->GetFd());
            // context_->lock_mgr_->lock_IX_on_table(context_->txn_, fh_->GetFd());
        }
        //由于还不涉及事务和索引因此只是删除对应的记录
        for(auto rid:rids_){
            auto rcd = fh_->get_record(rid, context_);
            //写日志
            DeleteLogRecord* delete_log=new DeleteLogRecord(context_->txn_->get_transaction_id(),*rcd,rid,tab_name_);
            delete_log->prev_lsn_ = context_->txn_->get_prev_lsn();
            context_->log_mgr_->add_log_to_buffer(delete_log);
            context_->txn_->set_prev_lsn(delete_log->lsn_);
            fh_->delete_record(rid,context_);
            //更新该页的最新修改的lsn用于重放日志
            fh_->update_page_lsn(rid.page_no, delete_log->lsn_);
            delete delete_log;

            for(IndexMeta index: tab_.indexes){
                int total_len = index.col_tot_len;
                char key[total_len];
                int offset = 0;
                for(ColMeta col: index.cols){
                    memcpy(key+offset, rcd->data+col.offset, col.len);
                    offset += col.len;
                }
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
                ih->delete_entry(key, context_->txn_);
            }

            
            

            //维护事务的write_set
            RmRecord write_rec(rcd->size);
            memcpy(write_rec.data,rcd->data,rcd->size);
            WriteRecord* write=new WriteRecord(WType::DELETE_TUPLE,tab_name_,rid,write_rec);
            context_->txn_->append_write_record(write);
            

            
        }
        
        return nullptr;
    }

    Rid &rid() override { return _abstract_rid; }
};