/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class InsertExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;                   // 表的元数据
    std::vector<Value> values_;     // 需要插入的数据
    RmFileHandle *fh_;              // 表的数据文件句柄
    std::string tab_name_;          // 表名称
    Rid rid_;                       // 插入的位置，由于系统默认插入时不指定位置，因此当前rid_在插入后才赋值
    SmManager *sm_manager_;

   public:
    InsertExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<Value> values, Context *context) {
        sm_manager_ = sm_manager;
        tab_ = sm_manager_->db_.get_table(tab_name);
        values_ = values;
        tab_name_ = tab_name;
        if (values.size() != tab_.cols.size()) {
            throw InvalidValueCountError();
        }
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        context_ = context;
    };

    std::unique_ptr<RmRecord> Next() override {
        
        // Make record buffer
        RmRecord rec(fh_->get_file_hdr().record_size);
        for (size_t i = 0; i < values_.size(); i++) {
            auto &col = tab_.cols[i];
            auto &val = values_[i];
            if(col.type==TYPE_FLOAT&&val.type==TYPE_INT){
                float temp=(float)val.int_val;
                val.set_float(temp);
            }
            if (col.type != val.type) {
                throw IncompatibleTypeError(coltype2str(col.type), coltype2str(val.type));
            }
            val.init_raw(col.len);
            memcpy(rec.data + col.offset, val.raw->data, col.len);
        }
        
        int index_num = tab_.indexes.size();
        char* key[index_num];
        //索引一致性检查
        for(size_t i = 0; i < tab_.indexes.size(); ++i) {
            auto& index = tab_.indexes[i];
            auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
            key[i] = new char[index.col_tot_len];
            int offset = 0;
            for(size_t j = 0; j < index.col_num; ++j) {
                memcpy(key[i] + offset, rec.data + index.cols[j].offset, index.cols[j].len);
                offset += index.cols[j].len;
            }
            std::vector<Rid> result;
            if(ih->get_value(key[i], &result, context_->txn_)){
                std::cout<<"1"<<std::endl;
                throw IndexEntryExistsError();
                // throw TransactionAbortException(context_->txn_->get_transaction_id(),AbortReason::INDEX_EXIST);
            }
        }
        if(context_!=nullptr){
            context_->lock_mgr_->lock_exclusive_on_table(context_->txn_, fh_->GetFd());
            // context_->lock_mgr_->lock_IX_on_table(context_->txn_, fh_->GetFd());
        }
        //插入记录
        rid_ = fh_->insert_record(rec.data, context_);
        //写日志
        InsertLogRecord* log=new InsertLogRecord(context_->txn_->get_transaction_id(),rec,rid_,tab_name_);
        log->prev_lsn_ = context_->txn_->get_prev_lsn();
        context_->log_mgr_->add_log_to_buffer(log);
        context_->txn_->set_prev_lsn(log->lsn_);
        //更新该页的最新修改的lsn用于重放日志
        fh_->update_page_lsn(rid_.page_no, log->lsn_);
        delete log;
        //索引维护
        for(size_t i = 0; i < tab_.indexes.size(); ++i){
            auto& index = tab_.indexes[i];
            auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
            ih->insert_entry(key[i], rid_, context_->txn_);
            // try{
            //     ih->insert_entry(key[i], rid_, context_->txn_);
            // }catch(RMDBError &e){
            //     if(dynamic_cast<IndexEntryExistsError*>(&e))
            //         {
            //             std::cout << "txn:thred_id: " << context_->txn_->get_thread_id() << " txn_id: " << context_->txn_->get_transaction_id()<<" start: "<<context_->txn_->get_start_ts() << std::endl;
            //             std::cerr << e.what()<< std::endl;
            //         }
            // }
            
        }
        //释放堆中的元素
        for(auto &it : key){
            delete [] it;
        }
        //事务维护
        RmRecord write_rec(rec.size);
        memcpy(write_rec.data,rec.data,rec.size);
        WriteRecord* write=new WriteRecord(WType::INSERT_TUPLE,tab_name_,rid_,write_rec);
        context_->txn_->append_write_record(write);
        
        
        
        return nullptr;
    }
    Rid &rid() override { return rid_; }
};