/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class NestedLoopJoinExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> left_;    // 左儿子节点（需要join的表）
    std::unique_ptr<AbstractExecutor> right_;   // 右儿子节点（需要join的表）
    size_t len_;                                // join后获得的每条记录的长度
    std::vector<ColMeta> cols_;                 // join后获得的记录的字段

    std::vector<Condition> fed_conds_;          // join条件
    bool isend;

   public:
    NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right, 
                            std::vector<Condition> conds) {
        left_ = std::move(left);
        right_ = std::move(right);
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto &col : right_cols) {
            col.offset += left_->tupleLen();
        }

        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        isend = false;
        fed_conds_ = std::move(conds);

    }

    void beginTuple() override {
        //从左边拿出来一个tuple
        left_->beginTuple();
        auto left_rec=left_->Next();
        //如果左边是个空表或者左边没有符合where条件的数据，则退出
        while(left_rec==nullptr&&!left_->is_end()){
            left_->nextTuple();
            left_rec=left_->Next();
        }
        if(left_->is_end()){
            return;
        }
        //然后从右表拿出来一个tuple
        right_->beginTuple();
        auto temp=left_next();
        if(temp==nullptr){
            return;
        }
        //可以优化：右表没有记录或者完全没有符合where条件的记录直接返回
    }

    //这个函数用于left取到下一个记录，right回到第一个记录
    RmRecord* left_next(){
        if(right_->is_end()){
            // static size_t count = 0;
            // count++;
            // std::cout << "NLJExecutor left next " << count << std::endl;
            //左表下一个记录
            left_->nextTuple();
            auto left_rec=left_->Next();
            while(left_rec==nullptr&&!left_->is_end()){
                left_->nextTuple();
                left_rec=left_->Next();
            }
            if(left_->is_end()){
                return nullptr;
            }
            //右表回到第一条记录
            right_->beginTuple();
        }
    }

    void nextTuple() override {
        // static size_t count = 0;
        // count++;
        // if (count % 100000 == 0) {
        //     std::cout << "NLJExecutor nextTuple" << std::endl;
        // }
        right_->nextTuple();
        auto temp=left_next();
        if(temp==nullptr){
            return;
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        // static size_t count = 0;
        // count++;
        // if (count % 100000 == 0) {
        //     // std::cout << "NLJExecutor Next" << std::endl;
        // }
        auto new_record=std::make_unique<RmRecord>(len_);
        auto left_rec=left_->Next();
        auto right_rec=right_->Next();
        if(left_rec==nullptr&&right_rec==nullptr){
            return nullptr;
        }
        memcpy(new_record->data,left_rec->data,left_rec->size);
        memcpy(new_record->data+left_rec->size,right_rec->data,right_rec->size);
        if(isequal(cols_,fed_conds_,new_record)){
            return new_record;
        }
        return nullptr;
    }

    //判断record中的数据是否符合conds中的data的数据,这个扫描不涉及连接查询
    bool isequal(const std::vector<ColMeta>& tab_cols,const std::vector<Condition>& conds,const std::unique_ptr<RmRecord>& rec){
        //循环比较一个记录的内容与每种where条件
        bool is_equal=true;
        for(const auto &cond:conds){
            //取出每条记录的列
            auto lhs_col=get_col(tab_cols,cond.lhs_col);
            //取出对应列的data
            char* lhs_data=rec->data+lhs_col->offset;
            ColType lhs_type=lhs_col->type;
            char* rhs_data;
            //取出右值
            if(cond.is_rhs_val){
                //右值把值存在记录里面了
                rhs_data=cond.rhs_val.raw->data;
            }else{
                auto rhs_col=get_col(tab_cols,cond.rhs_col);
                rhs_data=rec->data+rhs_col->offset;
                
            }
            
            //使用ix_compare函数进行对比
            const int col_len=lhs_col->len;
            int num=ix_compare(lhs_data,rhs_data,lhs_type,col_len);
            if(cond.op==OP_EQ){
                is_equal=(num==0)&&is_equal;
            }else if(cond.op==OP_GE){
                is_equal=(num>=0)&&is_equal;
            }else if(cond.op==OP_GT){
                is_equal=(num>0)&&is_equal;
            }else if(cond.op==OP_LE){
                is_equal=(num<=0)&&is_equal;
            }else if(cond.op==OP_LT){
                is_equal=(num<0)&&is_equal;
            }else{
                is_equal=(num!=0)&&is_equal;
            }
            if(is_equal==false){
                return is_equal;
            }
        }
        return is_equal;
    }
    Rid &rid() override { return _abstract_rid; }

    std::string getType() override { return "nestedloop_join"; };

    bool is_end() const override { return left_->is_end(); };

    ColMeta get_col_offset(const TabCol &target){
        for(int i=0;i<cols_.size();i++){
            if(cols_[i].name==target.col_name&&cols_[i].tab_name==target.tab_name){
                return cols_[i];
            }
        }
    }

    const std::vector<ColMeta> &cols() const override {
        return cols_;
    }

    size_t tupleLen() const override{ return len_; };
};