/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class ProjectionExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> prev_;        // 投影节点的儿子节点
    std::vector<ColMeta> cols_;                     // 需要投影的字段
    size_t len_;                                    // 字段总长度,在处理数据时可用于预先分配内存或计算缓冲区大小
    std::vector<size_t> sel_idxs_;                  
    // 含义: 保存了每个待投影列在上一个执行器的列集合中的索引位置。这有助于快速访问上游数据中对应的列数据。

   public:
    ProjectionExecutor(std::unique_ptr<AbstractExecutor> prev, const std::vector<TabCol> &sel_cols) {
        prev_ = std::move(prev);

        size_t curr_offset = 0;
        auto &prev_cols = prev_->cols();
        // if(auto x=std::dynamic_pointer_cast<AggExecutor>(prev_)){}
        for (auto &sel_col : sel_cols) {
            auto pos = get_col(prev_cols, sel_col);
            // 对于sel_cols中的每一项，调用get_col(prev_cols, sel_col)找到对应列在prev_->cols()中的位置。
            // get_col用于根据列名或其他标识从列集合中查找列。
            sel_idxs_.push_back(pos - prev_cols.begin());// 计算孩子节点中查询列的索引: 更新每一列的索引，放入sel_idxs_中
            auto col = *pos;
            col.offset = curr_offset;// 计算每一列的偏移量，累加到curr_offset，用于记录数据行中各列的布局。
            curr_offset += col.len;
            cols_.push_back(col);
        }
        len_ = curr_offset;
    }

    void beginTuple() override 
    {
        // projection算子作为根节点，需要调用下方的seq_scan算子
        prev_->beginTuple();
    }

    void nextTuple() override 
    {
        prev_->nextTuple();
    }

    std::unique_ptr<RmRecord> Next() override {
        // Next输出每个select需要查询的一行信息
        // 如果没有符合条件的返回空
        std::unique_ptr<RmRecord> tuple;
        tuple = prev_->Next();
        if(tuple == nullptr)
        {   //??
            return nullptr;
        }
        
        // 如果有符合条件的记录，遍历符合条件的所有记录
        auto record_ptr(std::make_unique<RmRecord>(int(len_)));
        ColMeta colmeta;
        for(size_t i = 0, j = 0; i < cols_.size(); i++)
        {   
            j = sel_idxs_[i];//拿到在孩子节点中的列序号
            colmeta = (prev_->cols())[j];
            memcpy(record_ptr->data + cols_[i].offset, tuple->data + colmeta.offset, colmeta.len);
        }
        return record_ptr;
    }

    bool is_end() const override{return prev_->is_end();};
    
    const std::vector<ColMeta> &cols() const override{return cols_;};
    
    Rid &rid() override { return _abstract_rid; }
};