/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SeqScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;              // 表的名称
    std::vector<Condition> conds_;      // scan的条件
    RmFileHandle *fh_;                  // 表的数据文件句柄
    std::vector<ColMeta> cols_;         // scan后生成的记录的字段
    size_t len_;                        // scan后生成的每条记录的长度
    std::vector<Condition> fed_conds_;  // 同conds_，两个字段相同

    Rid rid_;
    std::unique_ptr<RecScan> scan_;     // table_iterator

    SmManager *sm_manager_;

   public:
    SeqScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, Context *context,
        bool flag) {
        sm_manager_ = sm_manager;
        tab_name_ = std::move(tab_name);
        conds_ = std::move(conds);
        TabMeta &tab = sm_manager_->db_.get_table(tab_name_);
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab.cols;
        len_ = cols_.back().offset + cols_.back().len;

        context_ = context;

        fed_conds_ = conds_;
        has_index = flag;
    }

    void beginTuple() override {
        context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
        scan_=std::make_unique<RmScan>(fh_);
        for(;!scan_->is_end();scan_->next()){
            //通过数据扫描器拿到rid
            rid_=scan_->rid();
            //根据数据扫描器拿到具体的记录
            bool success=true;
            RmRecord* rec;
            std::unique_ptr<RmRecord> t;
            try{
                t=fh_->get_record(rid_,context_);
                // rec=(fh_->get_record(rid_,context_)).get();
            }catch(RecordNotFoundError& e){
                std::cout<<"record not found"<<std::endl;
                //没有拿到记录判断为false,这部分的异常处理没有想好
                success=false;
            }
            //已经取到了数据，那么将record中的data和conds中的data做对比
            if(success){
                bool is_equal=isequal(cols_,conds_,t);
                if(is_equal){
                    break;
                }
            }
        }
    }
    //判断record中的数据是否符合conds中的data的数据,这个扫描不涉及连接查询
    bool isequal(const std::vector<ColMeta>& tab_cols,const std::vector<Condition>& conds,const std::unique_ptr<RmRecord>& rec){
        if(context_!=nullptr){
            context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
            // context_->lock_mgr_->lock_IS_on_table(context_->txn_, fh_->GetFd());
        }
        //循环比较一个记录的内容与每种where条件
        bool is_equal=true;
        for(const auto& cond:conds){
            //取出每条记录的列
            auto lhs_col=get_col(tab_cols,cond.lhs_col);
            //取出对应列的data
            char* lhs_data=rec->data+lhs_col->offset;
            ColType lhs_type=lhs_col->type;
            char* rhs_data;
            //取出右值
            if(cond.is_rhs_val){
                //右值把值存在记录里面了
                rhs_data=cond.rhs_val.raw->data;
            }else{
                auto rhs_col=get_col(tab_cols,cond.rhs_col);
                rhs_data=rec->data+rhs_col->offset;
                
            }
            
            //使用ix_compare函数进行对比
            const int col_len=lhs_col->len;
            int num=ix_compare(lhs_data,rhs_data,lhs_type,col_len);
            if(cond.op==OP_EQ){
                is_equal=(num==0)&&is_equal;
            }else if(cond.op==OP_GE){
                is_equal=(num>=0)&&is_equal;
            }else if(cond.op==OP_GT){
                is_equal=(num>0)&&is_equal;
            }else if(cond.op==OP_LE){
                is_equal=(num<=0)&&is_equal;
            }else if(cond.op==OP_LT){
                is_equal=(num<0)&&is_equal;
            }else{
                is_equal=(num!=0)&&is_equal;
            }
            if(is_equal==false){
                return is_equal;
            }
        }
        return is_equal;
    }

    void nextTuple() override {
        for(scan_->next();!scan_->is_end();scan_->next()){
            //通过数据扫描器拿到rid
            rid_=scan_->rid();
            //根据数据扫描器拿到具体的记录
            bool success=true;
            RmRecord* rec;
            std::unique_ptr<RmRecord> t;
            try{
                t=fh_->get_record(rid_,context_);
                // rec=(fh_->get_record(rid_,context_)).get();
            }catch(RecordNotFoundError& e){
                std::cout<<"record not found"<<std::endl;
                //没有拿到记录判断为false,这部分的异常处理没有想好
                success=false;
            }
            //已经取到了数据，那么将record中的data和conds中的data做对比
            if(success){
                bool is_equal=isequal(cols_,conds_,t);
                if(is_equal){
                    break;
                }
            }
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        return fh_->get_record(rid_,context_);
    }

    std::string getType() override{return "SeqScan";};

    bool is_end() const override{return scan_->is_end();};

    Rid &rid() override { return rid_; }

    size_t tupleLen() const override{return len_;};

    const std::vector<ColMeta> &cols() const override{return cols_;};
    
    ColMeta get_col_offset(const TabCol &target){
        for(int i=0;i<cols_.size();i++){
            if(cols_[i].name==target.col_name&&cols_[i].tab_name==target.tab_name){
                return cols_[i];
            }
        }
    }
   
    std::string get_table_name(){
        return tab_name_;
    }

};