/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class UpdateExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;
    std::vector<Condition> conds_;
    RmFileHandle *fh_;
    std::vector<Rid> rids_;
    std::string tab_name_;
    std::vector<SetClause> set_clauses_;
    SmManager *sm_manager_;

   public:
    UpdateExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<SetClause> set_clauses,
                   std::vector<Condition> conds, std::vector<Rid> rids, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        set_clauses_ = set_clauses;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        context_ = context;
    }
    std::unique_ptr<RmRecord> Next() override {
        if(context_!=nullptr){  
            context_->lock_mgr_->lock_exclusive_on_table(context_->txn_, fh_->GetFd());
            // context_->lock_mgr_->lock_IX_on_table(context_->txn_, fh_->GetFd());
        }
        //存在类型转换的可能
        for (auto& set_caluse : set_clauses_){
            auto col_meta = tab_.get_col(set_caluse.lhs.col_name);
            if(col_meta->type == TYPE_FLOAT && set_caluse.rhs.type == TYPE_INT){
                float temp_value = (float)set_caluse.rhs.int_val;
                set_caluse.rhs.set_float(temp_value);
            }
            set_caluse.rhs.init_raw(col_meta->len);
        }

        for(auto &rid :rids_){
            auto rcd = fh_->get_record(rid, context_);
            //新建一个的记录对象来暂时存放修改后的记录数据
            RmRecord temp_rcd{rcd->size};
            RmRecord old_rcd{rcd->size};
            //先将需要更新的记录数据写入临时记录中
            memcpy(temp_rcd.data, rcd->data, rcd->size);
            memcpy(old_rcd.data, temp_rcd.data, rcd->size);
            //遍历set子句，开始修改临时记录的值
            for (auto& set_caluse : set_clauses_) {
                auto col_meta = tab_.get_col(set_caluse.lhs.col_name);
                if(set_caluse.is_add == false){
                    memcpy(temp_rcd.data + col_meta->offset, set_caluse.rhs.raw->data, col_meta->len);
                }
                else{
                    // 代表是update t1 set c1=c1+1;这种加法update语句
                    Value temp_val;
                    switch (col_meta->type)
                    {
                        case TYPE_FLOAT:
                            temp_val.set_float(*(float*)(temp_rcd.data + col_meta->offset) + set_caluse.rhs.float_val);
                            break;
                        case TYPE_INT:
                            temp_val.set_int(*(int*)(temp_rcd.data + col_meta->offset) + set_caluse.rhs.int_val);
                            break;
                    }
                    temp_val.init_raw(col_meta->len);
                    memcpy(temp_rcd.data + col_meta->offset, temp_val.raw->data, col_meta->len);
                }
            }
            
            
            int index_num = tab_.indexes.size();
            char *new_key[index_num];
            char *old_key[index_num];
            
            //对索引一致性检查
            for(int i = 0; i < index_num; i++){
                IndexMeta index = tab_.indexes[i];
                
                int total_len = index.col_tot_len;
                new_key[i] = new char[total_len];
                old_key[i] = new char[total_len];

                int offset = 0;
                for(ColMeta col: index.cols){
                    memcpy(old_key[i]+offset, old_rcd.data+col.offset, col.len);
                    memcpy(new_key[i]+offset, temp_rcd.data+col.offset, col.len);
                    offset += col.len;
                }
                if(memcmp(old_key[i], new_key[i], total_len) == 0){
                    continue;
                }
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
                std::vector<Rid> result;
                if(ih->get_value(new_key[i], &result, context_->txn_)){
                    // std::cout<<"在表"<<index.tab_name<<"的";
                    // for(auto it : index.cols){
                    //     std::cout<<" "<<it.name<<" ";
                    // }
                    // std::cout<<"索引上发生一致性冲突"<<std::endl;
                    throw IndexEntryExistsError();
                    // throw TransactionAbortException(context_->txn_->get_transaction_id(),AbortReason::INDEX_EXIST);
                }
            }
            //索引维护
            for(int i = 0; i < index_num; i++){
                IndexMeta index = tab_.indexes[i];
                int total_len = index.col_tot_len;
                
                if(memcmp(old_key[i], new_key[i], total_len) == 0){
                    continue;
                }
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
                ih->delete_entry(old_key[i], context_->txn_);
                ih->insert_entry(new_key[i], rid, context_->txn_);
            }
            //释放堆中的元素
            for(auto& it : new_key){
                delete [] it;
            }
            for(auto& it : old_key){
                delete [] it;
            }
            //写日志
            UpdateLogRecord* log=new UpdateLogRecord(context_->txn_->get_transaction_id(),temp_rcd,old_rcd,rid,tab_name_);
            log->prev_lsn_ = context_->txn_->get_prev_lsn();
            //context_->txn_->set_prev_lsn(context_->log_mgr_->add_log_to_buffer(log));
            context_->log_mgr_->add_log_to_buffer(log);
            context_->txn_->set_prev_lsn(log->lsn_);
            //更新该页的最新修改的lsn用于重放日志
            fh_->update_page_lsn(rid.page_no, log->lsn_);
            delete log;
            //真正修改记录
            fh_->update_record(rid, temp_rcd.data, context_);
            //事务维护
            //RmRecord write_rec(temp_rcd.size);
            //memcpy(write_rec.data,old_rcd.data,old_rcd.size);
            WriteRecord* write=new WriteRecord(WType::UPDATE_TUPLE,tab_name_,rid, old_rcd);
            context_->txn_->append_write_record(write);
            
        }
        
        return nullptr;
    }

    Rid &rid() override { return _abstract_rid; }
};