/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "ix_index_handle.h"

#include "ix_scan.h"

/**
 * @brief 在当前node中查找第一个>=target的key_idx
 *
 * @return key_idx，范围为[0,num_key)，如果返回的key_idx=num_key，则表示target大于最后一个key
 * @note 返回key index（同时也是rid index），作为slot no
 */
int IxNodeHandle::lower_bound(const char *target) const {
    // Todo:
    // 查找当前节点中第一个大于等于target的key，并返回key的位置给上层
    // 提示: 可以采用多种查找方式，如顺序遍历、二分查找等；使用ix_compare()函数进行比较

    // const int THRESHOLD = get_max_size(); // 设定阈值，可根据实际情况调整 
    const int THRESHOLD = -1;
    // 阈值待讨论lun
    int key_size = get_size();  //0代表没有key，1代表一个...，下边key_size等于1的时候通过get_key(0)能找到第一个key
    if(key_size < THRESHOLD)
    {   // 如果节点个数不多的时候就顺序遍历
        for(int i = 0; i < key_size; i++)
        {   
            // 遍历当前节点中每个key，进行比较，如果key大于等于target就返回
            int cmp_result = ix_compare(get_key(i), target, file_hdr->col_types_, file_hdr->col_lens_);
            if(cmp_result >= 0)
            {
                return i;
            }
        }
    }
    else
    {   // 节点个数较多时采用二分查找
        int left = 0;
        int right = key_size;
        while (left < right) 
        {
            int mid = left + (right - left) / 2;
            int cmp_result = ix_compare(get_key(mid), target, file_hdr->col_types_, file_hdr->col_lens_);
            if (cmp_result >= 0) {
                right = mid;
            } else{
                left = mid + 1;
            }
        }
        return left;
    }

    return key_size;    // 否则没有比target大的key，表示target大于最后一个key，返回num_key
}

/**
 * @brief 在当前node中查找第一个>target的key_idx
 *
 * @return key_idx，范围为[1,num_key)，如果返回的key_idx=num_key，则表示target大于等于最后一个key
 * @note 注意此处的范围从1开始
 */
// 这个函数和lower_bound()不同点：1.查找严格大于的key_idx 2.返回key_idx=num_key表示target大于等于最后一个key（不是大于）3.范围从1开始
// 二分查找待验证lun
int IxNodeHandle::upper_bound(const char *target) const {
    // Todo:
    // 查找当前节点中第一个大于target的key，并返回key的位置给上层
    // 提示: 可以采用多种查找方式：顺序遍历、二分查找等；使用ix_compare()函数进行比较
    // const int THRESHOLD = get_max_size(); // 设定阈值，可根据实际情况调整 
    const int THRESHOLD = -1;
    // 阈值待讨论lun
    int key_size = get_size();
    if(key_size < THRESHOLD)
    {   
        // 如果节点个数不多的时候就顺序遍历
        for(int i = 1; i < key_size; i++)
        {   
            // 遍历当前节点中每个key，进行比较，如果key大于target就返回
            int cmp_result = ix_compare(get_key(i), target, file_hdr->col_types_, file_hdr->col_lens_);
            if(cmp_result > 0)
            {
                return i;
            }
        }
        // 否则没有比target大的key，表示target大于最后一个key，返回num_key
        return key_size;
    }
    else
    {   // 节点个数较多时采用二分查找
        int left = 1;
        int right = key_size;
        while (left < right)
        {
            int mid = left + (right - left) / 2;
            int cmp_result = ix_compare(get_key(mid), target, file_hdr->col_types_, file_hdr->col_lens_);
            if (cmp_result > 0) {
                right = mid;
            } else {
                left = mid + 1; // 如果找到相等或者小于的，都需要继续在右侧查找大于target的key
            }
        }
        return left;
    }  
}

/**
 * @brief 用于叶子结点根据key来查找该结点中的键值对
 * 值value作为传出参数，函数返回是否查找成功
 *
 * @param key 目标key
 * @param[out] value 传出参数，目标key对应的Rid
 * @return 目标key是否存在
 */
bool IxNodeHandle::leaf_lookup(const char *key, Rid **value) {
    // Todo:
    // 1. 在叶子节点中获取目标key所在位置
    // 2. 判断目标key是否存在
    // 3. 如果存在，获取key对应的Rid，并赋值给传出参数value
    // 提示：可以调用lower_bound()和get_rid()函数。
    int key_idx = lower_bound(key);
    int key_size = get_size();
    if(key_idx < key_size)
    {
        // 如果key_idx < key_size说明lower_bound返回的有>=target的key_idx
        if(0 == ix_compare(get_key(key_idx), key, file_hdr->col_types_, file_hdr->col_lens_))
        {
            *value = get_rid(key_idx);  // 等于key说明找到了，赋值value
            return true; 
        }
    }

    // 1.比节点中所有的key都大 2.lower_bound找的不是=的，是>的
    return false;   // lower_bound()返回num_size或者返回的对应的结果key>参数Key,都返回false

}

/**
 * 用于内部结点（非叶子节点）查找目标key所在的孩子结点（子树）
 * @param key 目标key
 * @return page_id_t 目标key所在的孩子节点（子树）的存储页面编号
 */
page_id_t IxNodeHandle::internal_lookup(const char *key) {
    // Todo:
    // 1. 查找当前非叶子节点中目标key所在孩子节点（子树）的位置
    // 2. 获取该孩子节点（子树）所在页面的编号
    // 3. 返回页面编号

    // 0 1 3 4     upper_bound lower_bound 
    // -1  -> 0 [0]   [1]         [0]
    // 0   -> 0 [0]   [1]         [0]
    // 1   -> 1 [1]   [2]         [1]
    // 2   -> 1 [1]   [2]         [2]
    // 5   -> 4 [3]   [4]         [4]

    // // 这里不应该判断叶子节点中是否存在对应key，仅仅返回如果有key应该在哪个叶子节点的页面号即可
    // int child_idx = lower_bound(key);
    // // lun这里有问题，如果找的key最小，child_idx=0里边也都比它大
    // if (child_idx == 0 ||
    // (0 == ix_compare(get_key(child_idx), key, file_hdr->col_types_, file_hdr->col_lens_) &&
    // child_idx != get_size())) 
    // {
    //     // 如果=0说明目标key小于当前节点的最小key，应选择第一个孩子节点
    //     // 如果key字符串 == get_key(child_idx)字符串，说明内节点的其中一个key刚好为要找的key，它的value指向的子树第一个节点就是要找的节点
    //     // 无操作
    // } 
    // else 
    // {   
    //     // 目标key位于child_idx-1的子树：lower_bound找的大，或者child_idx == get_size()
    //     child_idx -= 1;  // 选择前一个孩子
    // }

    // return value_at(child_idx);  // 获取孩子节点的页面编号

    int child_idx = upper_bound(key);
    child_idx -= 1;
    return value_at(child_idx);  // 获取孩子节点的页面编号
}

/**
 * @brief 在指定位置插入n个连续的键值对
 * 将key的前n位插入到原来keys中的pos位置；将rid的前n位插入到原来rids中的pos位置
 *
 * @param pos 要插入键值对的位置
 * @param (key, rid) 连续键值对的起始地址，也就是第一个键值对，可以通过(key, rid)来获取n个键值对
 * @param n 键值对数量
 * @note [0,pos)           [pos,num_key)
 *                            key_slot
 *                            /      \
 *                           /        \
 *       [0,pos)     [pos,pos+n)   [pos+n,num_key+n)
 *                      key           key_slot
 */
void IxNodeHandle::insert_pairs(int pos, const char *key, const Rid *rid, int n) {
    // Todo:
    // 1. 判断pos的合法性
    // 2. 通过key获取n个连续键值对的key值，并把n个key值插入到pos位置
    // 3. 通过rid获取n个连续键值对的rid值，并把n个rid值插入到pos位置
    // 4. 更新当前节点的键数量

    // lun这里边界需要讨论，如果插入的位置不是覆盖，那么1.pos<0 2.过大 3.是否要和之前的num_key连续 判断非法
    int key_size = get_size();  // pos∈[0,key_size]，如果pos=key_size就是在末尾追加插入
    int move_num = key_size - pos;    //移动的kv数量
    // 插入判断pos需要连续，所以不能> key_size，但是可以等于（也就是在末尾插入）
    if(pos < 0 || pos > key_size)
    {
        // 这里没有对应的异常处理lun
        return;
    }
    else
    {   
        if(move_num > 0)
        {
            // 说明要移动原有的键值对(key-value)位置
            memmove(keys + pos * file_hdr->col_tot_len_ + n * file_hdr->col_tot_len_, keys + pos * file_hdr->col_tot_len_, move_num * file_hdr->col_tot_len_);
            // lun这里用不用set_rid?
            for(int i = 0; i < move_num; i++)
            {
                rids[key_size-1-i+n] = rids[key_size-1-i];
            }
        }
        // 插入应该插入的kv
        for(int i = 0; i < n; i++)
        {
            memcpy(keys + pos * file_hdr->col_tot_len_ + i * file_hdr->col_tot_len_, key + i * file_hdr->col_tot_len_, file_hdr->col_tot_len_);
            rids[pos + i] = rid[i];
            // lun这里要不要不用for，一次赋值？
        }
        set_size(get_size() + n);
    }
}

/**
 * @brief 用于在结点中插入单个键值对。
 * 函数返回插入后的键值对数量
 *
 * @param (key, value) 要插入的键值对
 * @return int 键值对数量
 */
int IxNodeHandle::insert(const char *key, const Rid &value, int &insert_idx) {
    // Todo:
    // 1. 查找要插入的键值对应该插入到当前节点的哪个位置
    // 2. 如果key重复则不插入
    // 3. 如果key不重复则插入键值对
    // 4. 返回完成插入操作之后的键值对数量
    insert_idx = lower_bound(key);
    int key_num = get_size();

    if(insert_idx < key_num && 0 == ix_compare(get_key(insert_idx), key, file_hdr->col_types_, file_hdr->col_lens_))
    {   
        return key_num; // 此条件下key重复，返回原本的kv数量
    }
    else
    {   
        const Rid *rid_ptr = &value;
        insert_pairs(insert_idx, key, rid_ptr, 1);
        // 调用insert_pairs()之后num_key在函数内部增加，直接返回就行
        return get_size();
    }
}

/**
 * @brief 用于在结点中的指定位置删除单个键值对
 *
 * @param pos 要删除键值对的位置
 */
void IxNodeHandle::erase_pair(int pos) {
    // Todo:
    // 1. 删除该位置的key
    // 2. 删除该位置的rid
    // 3. 更新结点的键值对数量
    int key_size = get_size();  // pos∈[0,key_size)
    // lun这里要不要int key_size还是每次用get_size()
    if(pos < 0 || pos >= key_size)
    {
        // 这里没有对应的异常处理lun
        return;
    }
    int move_num = key_size - 1 - pos;    //需要移动的数量
    if(move_num > 0)
    {
        memmove(keys + pos * file_hdr->col_tot_len_, keys + (pos + 1) * file_hdr->col_tot_len_, move_num * file_hdr->col_tot_len_);
        // lun这里用memmove还是for好？
        for(int i = 0; i < move_num; i++)
        {
            set_rid(pos+i, *(get_rid(pos+i+1)));
        }
    }
    // 清除最后一个pair：1.移动后最后一个需要清除 2.pos=key_size-1也需要清除
    // 不需要单独修改数据，把key_size-1后边的无效就行了
    set_size(key_size - 1);
}

/**
 * @brief 用于在结点中删除指定key的键值对。函数返回删除后的键值对数量
 *
 * @param key 要删除的键值对key值
 * @return 完成删除操作后的键值对数量
 */
int IxNodeHandle::remove(const char *key) {
    // Todo:
    // 1. 查找要删除键值对的位置
    // 2. 如果要删除的键值对存在，删除键值对
    // 3. 返回完成删除操作后的键值对数量
    int key_idx = lower_bound(key);
    int key_num = get_size();
    if(key_idx < key_num && 0 == ix_compare(get_key(key_idx), key, file_hdr->col_types_, file_hdr->col_lens_))
    {   
        erase_pair(key_idx);
        return get_size(); // erase_pair中删除后已经将num_key-1，直接返回即可
    }
    else
    {
        // lun异常抛出
        return key_num;//否则返回当前pair数量（未删除的时候）
    }
}

IxIndexHandle::IxIndexHandle(DiskManager *disk_manager, BufferPoolManager *buffer_pool_manager, int fd)
    : disk_manager_(disk_manager), buffer_pool_manager_(buffer_pool_manager), fd_(fd) {
    // init file_hdr_
    disk_manager_->read_page(fd, IX_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
    char* buf = new char[PAGE_SIZE];
    memset(buf, 0, PAGE_SIZE);
    disk_manager_->read_page(fd, IX_FILE_HDR_PAGE, buf, PAGE_SIZE);
    file_hdr_ = new IxFileHdr();
    file_hdr_->deserialize(buf);
    delete[] buf;
    
    // disk_manager管理的fd对应的文件中，设置从file_hdr_->num_pages开始分配page_no
    int now_page_no = disk_manager_->get_fd2pageno(fd);
    disk_manager_->set_fd2pageno(fd, now_page_no + 1);
}

/**
 * @brief 用于查找指定键所在的叶子结点
 * @param key 要查找的目标key值
 * @param operation 查找到目标键值对后要进行的操作类型
 * @param transaction 事务参数，如果不需要则默认传入nullptr
 * @return [leaf node] and [root_is_latched] 返回目标叶子结点以及根结点是否加锁
 * @note need to Unlatch and unpin the leaf node outside!
 * 注意：用了FindLeafPage之后一定要unlatch叶结点，否则下次latch该结点会堵塞！
 */
std::pair<IxNodeHandle *, bool> IxIndexHandle::find_leaf_page(const char *key, Operation operation,
                                                            Transaction *transaction, bool find_first) {
    // Todo:
    // 1. 获取根节点
    // 2. 从根节点开始不断向下查找目标key
    // 3. 找到包含该key值的叶子结点停止查找，并返回叶子节点

    // lun find_first是干什么的？
    // bool root_is_latched = false;
    IxNodeHandle * curr_node = nullptr;
    
    // 获取根节点
    // std::lock_guard<std::mutex> latch(root_latch_);
    // root_is_latched = true;
    page_id_t root_page_no = file_hdr_->root_page_;
    if (root_page_no == IX_NO_PAGE) {
        // lun异常处理
        return std::make_pair(nullptr, false);  // 空树
    }
    curr_node = fetch_node(root_page_no);
    // 自动释放root_latch_

    while (!curr_node->is_leaf_page())
    {   
        page_id_t child_page_id = curr_node->internal_lookup(key);
        buffer_pool_manager_->unpin_page(curr_node->get_page_id(), false);
        delete curr_node;   // fetch_node中有new
        curr_node = fetch_node(child_page_id);
    }
    // 此时curr_node已经为叶子节点，如果要找的key不存在(FIND和INSERT)，key的值不同情况此时curr_node的赋值情况如下：
    // 1.最小：返回最左侧叶子节点 
    // 2.中间：返回比找的key小的key所在的叶子结点page_no 
    // 3.最大：返回最右侧叶子节点
    // 如果是DELETE则正常返回实际存在的key所在的叶子节点page_no
    
    switch (operation)
    {
        case Operation::FIND:{
            break;  
        }
        case Operation::INSERT:
        case Operation::DELETE:
            break;
    }
    // 因为要返回叶子节点的指针，所以说明还在用，先不unpin_page
    // buffer_pool_manager_->unpin_page(curr_node->get_page_id(), false);
    return std::make_pair(curr_node, true); // lun改锁
}

/**
 * @brief 用于查找指定键在叶子结点中的对应的值result
 *
 * @param key 查找的目标key值
 * @param result 用于存放结果的容器
 * @param transaction 事务指针
 * @return bool 返回目标键值对是否存在
 */
bool IxIndexHandle::get_value(const char *key, std::vector<Rid> *result, Transaction *transaction) {
    // Todo:
    // 1. 获取目标key值所在的叶子结点
    // 2. 在叶子节点中查找目标key值的位置，并读取key对应的rid
    // 3. 把rid存入result参数中
    // 提示：使用完buffer_pool提供的page之后，记得unpin page；记得处理并发的上锁

    auto leaf_node_pair = find_leaf_page(key, Operation::FIND, transaction);
    IxNodeHandle *leaf_node = leaf_node_pair.first;
    Rid rid({INVALID_PAGE_ID, -1});  // 创建一个 Rid 对象
    Rid *rid_ptr = &rid;           // 获取 Rid 对象的地址
    // Rid *rid_ptr = new Rid({INVALID_PAGE_ID, -1});   // 准备一个Rid指针的引用，用于传递给leaf_lookup
    
    if(!(leaf_node->leaf_lookup(key, &rid_ptr)))
    {   // 如果没找到目标键值对
        buffer_pool_manager_->unpin_page(leaf_node->get_page_id(), false);
        delete leaf_node;
        return false;   
    }
    else
    {   // 如果找到目标键值对
        buffer_pool_manager_->unpin_page(leaf_node->get_page_id(), false);
        result->push_back(*rid_ptr);
        delete leaf_node;
        return true;
    }
}

/**
 * @brief  将传入的一个node拆分(Split)成两个结点，在node的右边生成一个新结点new node
 * @param node 需要拆分的结点
 * @return 拆分得到的new_node
 * @note need to unpin the new node outside
 * 注意：本函数执行完毕后，原node和new node都需要在函数外面进行unpin
 */
IxNodeHandle *IxIndexHandle::split(IxNodeHandle *node) {
    // Todo:
    // 1. 将原结点的键值对平均分配，右半部分分裂为新的右兄弟结点
    //    需要初始化新节点的page_hdr内容
    // 2. 如果新的右兄弟结点是叶子结点，更新新旧节点的prev_leaf和next_leaf指针
    //    为新节点分配键值对，更新旧节点的键值对数记录
    // 3. 如果新的右兄弟结点不是叶子结点，更新该结点的所有孩子结点的父节点信息(使用IxIndexHandle::maintain_child())
    int mid = (node->get_size()) / 2;
    // mid左侧都给旧节点，如果node最多有偶数4个key,插入后key_size暂时为5，mid=2,右侧insert_num=5-2=3
    // 如果node最多奇数5个key，插入后key_size暂时为6，mid=3,右侧insert_num=6-3=3
    int insert_num = node->get_size() - mid;
    IxNodeHandle *new_rsibling_node =  create_node();

    // page_hdr中next_free_page_no未使用，应该可以不初始化，但是仿照ix_manager.h中初始化为-1
    new_rsibling_node->page_hdr->next_free_page_no = IX_NO_PAGE;
    new_rsibling_node->page_hdr->num_key = 0;
    new_rsibling_node->page_hdr->is_leaf = node->is_leaf_page();
    new_rsibling_node->set_parent_page_no(node->get_parent_page_no());  //这里父节点先这样，如果父节点分裂会在其他地方改
    // prev_leaf和next_leaf初始化
    if(new_rsibling_node->is_leaf_page())
    {
        // 只有叶子节点才有前后节点，内节点不使用这两个属性
        // 叶子节点是IX_LEAF_HEADER_PAGE开始的循环双向链表，每个叶子节点都有前后叶子节点
        /*
            原本：      
                parent_node-?
                    |       |
                  node <-> old_rsibling_node
            现在：
                   parent_node -------------------- ?
                    |       |                       |
                  node <-> new_rsibling_node <-> old_rsibling_node
        */
        new_rsibling_node->set_prev_leaf(node->get_page_no());
        new_rsibling_node->set_next_leaf(node->get_next_leaf());
        // std::cout<<"test";
        IxNodeHandle *old_rsibling_node =  fetch_node(node->get_next_leaf());   //先拿到，下一行再变next_leaf
        node->set_next_leaf(new_rsibling_node->get_page_no());
        old_rsibling_node->set_prev_leaf(new_rsibling_node->get_page_no());
        buffer_pool_manager_->unpin_page(old_rsibling_node->get_page_id(), true);   // old_rsibling_node改完prev_leaf就用不到了
        delete old_rsibling_node;
        // 为新节点分配键值对，删除旧节点被分出来的键值对(这里不用真的删除数据，把num_key值修改，多余的为非法访问即可)
        new_rsibling_node->insert_pairs(0, node->get_key(mid), node->get_rid(mid), insert_num); //把node右半部分的值插入到新右兄弟节点中
        node->set_size(mid);    // 因为node肯定是从结尾分走的，把num_key改了就行，不用调用node->erase_pair();
    }
    else
    {   
        new_rsibling_node->insert_pairs(0, node->get_key(mid), node->get_rid(mid), insert_num); //把node右半部分的值插入到新右兄弟节点中
        node->set_size(mid);    // 因为node肯定是从结尾分走的，把num_key改了就行，不用调用node->erase_pair();
        for(int i = 0; i < insert_num; i++)
        {   
            maintain_child(new_rsibling_node, i);
        }
    }

    return new_rsibling_node;
}

/**
 * @brief Insert key & value pair into internal page after split
 * 拆分(Split)后，向上找到old_node的父结点
 * 将new_node的第一个key插入到父结点，其位置在 父结点指向old_node的孩子指针 之后
 * 如果插入后>=maxsize，则必须继续拆分父结点，然后在其父结点的父结点再插入，即需要递归
 * 直到找到的old_node为根结点时，结束递归（此时将会新建一个根R，关键字为key，old_node和new_node为其孩子）
 *
 * @param (old_node, new_node) 原结点为old_node，old_node被分裂之后产生了新的右兄弟结点new_node
 * @param key 要插入parent的key
 * @note 一个结点插入了键值对之后需要分裂，分裂后左半部分的键值对保留在原结点，在参数中称为old_node，
 * 右半部分的键值对分裂为新的右兄弟节点，在参数中称为new_node（参考Split函数来理解old_node和new_node）
 * @note 本函数执行完毕后，new node和old node都需要在函数外面进行unpin
 */
void IxIndexHandle::insert_into_parent(IxNodeHandle *old_node, const char *key, IxNodeHandle *new_node,
                                     Transaction *transaction) {
    // Todo:
    // 1. 分裂前的结点（原结点, old_node）是否为根结点，如果为根结点需要分配新的root
    // 2. 获取原结点（old_node）的父亲结点
    // 3. 获取key对应的rid，并将(key, rid)插入到父亲结点
    // 4. 如果父亲结点仍需要继续分裂，则进行递归插入
    // 提示：记得unpin page
    IxNodeHandle *parent_node = nullptr;
    if(old_node->is_root_page())
    {   
        // 根节点需要插入两个（由于这个B+树的结构问题不是key_size+1=value_size(也就是指针数量或者rid_size),而是它们都相等）
        IxNodeHandle *new_root_node = create_node();
        new_root_node->page_hdr->next_free_page_no = IX_NO_PAGE;
        new_root_node->page_hdr->num_key = 0;
        new_root_node->page_hdr->is_leaf = false;
        new_root_node->set_parent_page_no(old_node->get_parent_page_no());

        file_hdr_->root_page_ = new_root_node->get_page_no();

        parent_node = new_root_node;
        parent_node->insert_pair(0, old_node->get_key(0), {old_node->get_page_no(), -1});
        parent_node->insert_pair(1, key, {new_node->get_page_no(), -1});

        // 修改父节点时间：
        // 1.叶子节点分裂的新节点父节点为分裂前的父节点 
        // 2.叶子节点分裂后进行insert_into_parent，如果父节点也需要分裂，分裂后的新父节点对应的孩子节点的父节点修改为新父节点
        // 3.到根节点分裂，需要再创建一个根节点，将这三个节点的父子关系进行分配
        old_node->page_hdr->parent = new_root_node->get_page_no(); 
        new_node->page_hdr->parent = new_root_node->get_page_no();
    }
    else
    {
        parent_node = fetch_node(old_node->get_parent_page_no());
        parent_node->insert_pair(parent_node->find_child(old_node) + 1, key, {new_node->get_page_no(), -1});
    }
    
    if(parent_node->get_size() >= parent_node->get_max_size())
    {
        IxNodeHandle *new_split_node = split(parent_node);
        insert_into_parent(parent_node, new_split_node->get_key(0), new_split_node, transaction);
        buffer_pool_manager_->unpin_page(parent_node->get_page_id(), true); 
        delete parent_node;
        buffer_pool_manager_->unpin_page(new_split_node->get_page_id(), true); 
        delete new_split_node;
    }
    else
    {
        buffer_pool_manager_->unpin_page(parent_node->get_page_id(), true); 
        delete parent_node;
    }
    // delete parent_node;
}

/**
 * @brief 将指定键值对插入到B+树中
 * @param (key, value) 要插入的键值对
 * @param transaction 事务指针
 * @return page_id_t 插入到的叶结点的page_no
 */
// 插入重复在内部会抛出异常
page_id_t IxIndexHandle::insert_entry(const char *key, const Rid &value, Transaction *transaction) {
    // Todo:
    // 1. 查找key值应该插入到哪个叶子节点
    // 2. 在该叶子节点中插入键值对
    // 3. 如果结点已满，分裂结点，并把新结点的相关信息插入父节点
    // 提示：记得unpin page；若当前叶子节点是最右叶子节点，则需要更新file_hdr_.last_leaf；记得处理并发的上锁
    // std::lock_guard<std::mutex> latch(root_latch_);
    std::scoped_lock lock{root_latch_};
    page_id_t insert_pageid = -1;
    auto leaf_page_pair = find_leaf_page(key, Operation::INSERT, transaction);
    IxNodeHandle *leaf_node = leaf_page_pair.first;
    int old_key_num = leaf_node->get_size();
    if(leaf_node == nullptr)
    {   
        buffer_pool_manager_->unpin_page(leaf_node->get_page_id(), false); 
        throw IndexEntryNotFoundError();
    }
    int insert_idx = -1;
    leaf_node->insert(key, value, insert_idx);
    // 下边需要更改父节点的值说明插入了leaf_node中的最小值（比现有的更小）
    // 而如果要插入一个节点的最小值，找到的leaf_node除了first_leaf，都是找一个大于这个key的最小节点，然后取这个节点的前一个节点
    // 如果是第二种情况就是插入一个节点的最大值，这有说明不用改parent中的key，所以只能是插入了所有数据的最小值才对maintain_parent(leaf_node)有实际作用
    if(insert_idx == 0)
    {
        maintain_parent(leaf_node); 
    }
    if(old_key_num == leaf_node->get_size())
    {   // 没有插入进去，代表插入的是重复的键值，由于为unique索引，所以抛出异常
        buffer_pool_manager_->unpin_page(leaf_node->get_page_id(), false); 
        delete leaf_node;
        // throw IndexEntryExistsError();
        throw TransactionAbortException(transaction->get_transaction_id(),AbortReason::INDEX_EXIST);
    }
    // 到这里说明已经正常插入
    if(leaf_node->get_size() >= leaf_node->get_max_size())
    {
        IxNodeHandle *new_split_node = split(leaf_node);
        if(leaf_node->get_page_no() == file_hdr_->last_leaf_)
        {
            file_hdr_->last_leaf_ = new_split_node->get_page_no();
        }
        insert_into_parent(leaf_node, new_split_node->get_key(0), new_split_node, transaction);
        // 判断插入的键值对在旧节点或者分裂的新节点中
        insert_pageid = find_leaf_page(key, Operation::FIND, transaction).first->get_page_no();
        // lun这里find_leaf_page会不会有泄漏？
        buffer_pool_manager_->unpin_page(leaf_node->get_page_id(), true); 
        delete leaf_node;
        buffer_pool_manager_->unpin_page(new_split_node->get_page_id(), true); 
        delete new_split_node;
    }
    else
    {   
        insert_pageid = leaf_node->get_page_no();   // 没分裂肯定就在leaf_node中
        buffer_pool_manager_->unpin_page(leaf_node->get_page_id(), true); 
        delete leaf_node;
    }
    // delete leaf_node;
    return insert_pageid;
}

/**
 * @brief 用于删除B+树中含有指定key的键值对
 * @param key 要删除的key值
 * @param transaction 事务指针
 */
bool IxIndexHandle::delete_entry(const char *key, Transaction *transaction) {
    //根节点锁
    std::scoped_lock lock{root_latch_};
    // Todo:
    // 1. 获取该键值对所在的叶子结点
    std::pair<IxNodeHandle * , bool> node_pair = find_leaf_page(key, Operation::DELETE, nullptr);
    IxNodeHandle *node = node_pair.first;
    //Iid iid = lower_bound(key);
    // 2. 在该叶子结点中删除键值对
    int origin_key_num = node->get_size();
    int afer_key_num = node->remove(key);
    //如果删除失败
    if(origin_key_num == afer_key_num){
        buffer_pool_manager_->unpin_page(node->get_page_id(), false);
        delete node;
        return false;
    }

    // 3. 如果删除成功需要调用CoalesceOrRedistribute来进行合并或重分配操作，并根据函数返回结果判断是否有结点需要删除
    bool need_delete = coalesce_or_redistribute(node, nullptr, &(node_pair.second));
    
    // (事务还不用管)4. 如果需要并发，并且需要删除叶子结点，则需要在事务的delete_page_set中添加删除结点的对应页面；记得处理并发的上锁

    buffer_pool_manager_->unpin_page(node->get_page_id(), true);
    delete node;
    return true;
}

/**
 * @brief 用于处理合并和重分配的逻辑，用于删除键值对后调用
 *
 * @param node 执行完删除操作的结点
 * @param transaction 事务指针
 * @param root_is_latched 传出参数：根节点是否上锁，用于并发操作
 * @return 是否需要删除结点
 * @note User needs to first find the sibling of input page.
 * If sibling's size + input page's size >= 2 * page's minsize, then redistribute.
 * Otherwise, merge(Coalesce).
 */
bool IxIndexHandle::coalesce_or_redistribute(IxNodeHandle *node, Transaction *transaction, bool *root_is_latched) {
    // Todo:
    // 1. 判断node结点是否为根节点
    if(node->is_root_page()){
        //    1.1 如果是根节点，需要调用AdjustRoot() 函数来进行处理，返回根节点是否需要被删除
        return adjust_root(node);
    }
        //    1.2 如果不是根节点，并且不需要执行合并或重分配操作，则直接返回false，否则执行2
    int node_size = node->get_size();
	int min_size = node->get_min_size();
    if(node_size >= min_size){
        return false;
    }
    // 2. 获取node结点的父亲结点
    page_id_t parent_page = node->get_parent_page_no();
    IxNodeHandle *parent_node = fetch_node(parent_page);
    // 3. 寻找node结点的兄弟结点（优先选取前驱结点）
    int i_no = parent_node->find_child(node);
    assert(i_no >= 0);
    page_id_t brother_page;
    if(i_no > 0){
        brother_page = parent_node->value_at(i_no-1);
    }
    else{
        brother_page = parent_node->value_at(i_no+1);
    }
    IxNodeHandle *brother_node = fetch_node(brother_page);
    // 4. 如果node结点和兄弟结点的键值对数量之和，能够支撑两个B+树结点（即node.size+neighbor.size >=
    // NodeMinSize*2)，则只需要重新分配键值对（调用Redistribute函数）
    PageId parent_page_id = parent_node->get_page_id();
    PageId borther_page_id = brother_node->get_page_id();
    bool need_delete;    
    if(brother_node->get_size() + node->get_size() >= min_size * 2){
        redistribute(brother_node, node, parent_node, i_no);
        buffer_pool_manager_->unpin_page(parent_page_id, true);
        delete parent_node;
        buffer_pool_manager_->unpin_page(borther_page_id, true);
        delete brother_node;
        need_delete = false;
    }
    // 5. 如果不满足上述条件，则需要合并两个结点，将右边的结点合并到左边的结点（调用Coalesce函数）
    else{
        need_delete = coalesce(brother_node, node, parent_node, i_no, transaction, root_is_latched);
        buffer_pool_manager_->unpin_page(parent_page_id, true);
        delete parent_node;
        buffer_pool_manager_->unpin_page(borther_page_id, true);
        delete brother_node;
    }
    return need_delete;
}

/**
 * @brief 用于当根结点被删除了一个键值对之后的处理
 * @param old_root_node 原根节点
 * @return bool 根结点是否需要被删除
 * @note size of root page can be less than min size and this method is only called within coalesce_or_redistribute()
 */
bool IxIndexHandle::adjust_root(IxNodeHandle *old_root_node) {
    // Todo:
    // 1. 如果old_root_node是内部结点，并且大小为1，则直接把它的孩子更新成新的根结点
    if(!old_root_node->is_leaf_page() && old_root_node->get_size() == 1){
        page_id_t childpage_no = old_root_node->value_at(0);
        IxNodeHandle *node = fetch_node(childpage_no);
        node->set_parent_page_no(INVALID_PAGE_ID);
        update_root_page_no(childpage_no);
        //仅仅是file_hdr_->num_pages_--？？
        release_node_handle(*old_root_node);
        //在缓冲区中删除原根节点所在的页,被pin住？
        buffer_pool_manager_->delete_page(old_root_node->get_page_id());
        buffer_pool_manager_->unpin_page(node->get_page_id(), true);
        delete node;
        return true;
    }
    // 2. 如果old_root_node是叶结点，且大小为0，则直接更新root page
    else if(old_root_node->is_leaf_page() && old_root_node->get_size() == 0){
        release_node_handle(*old_root_node);
        update_root_page_no(INVALID_PAGE_ID);
        return true;
    }
    
    // 3. 除了上述两种情况，不需要进行操作

    return false;
}

/**
 * @brief 重新分配node和兄弟结点neighbor_node的键值对
 * Redistribute key & value pairs from one page to its sibling page. If index == 0, move sibling page's first key
 * & value pair into end of input "node", otherwise move sibling page's last key & value pair into head of input "node".
 *
 * @param neighbor_node sibling page of input "node"
 * @param node input from method coalesceOrRedistribute()
 * @param parent the parent of "node" and "neighbor_node"
 * @param index node在parent中的rid_idx
 * @note node是之前刚被删除过一个key的结点
 * index=0，则neighbor是node后继结点，表示：node(left)      neighbor(right)
 * index>0，则neighbor是node前驱结点，表示：neighbor(left)  node(right)
 * 注意更新parent结点的相关kv对
 */
void IxIndexHandle::redistribute(IxNodeHandle *neighbor_node, IxNodeHandle *node, IxNodeHandle *parent, int index) {
    // Todo:
    // 1. 通过index判断neighbor_node是否为node的前驱结点
    if(index == 0){
        node->insert_pair(node->get_size(), neighbor_node->get_key(0), *(neighbor_node->get_rid(0)));
        neighbor_node->erase_pair(0);
        maintain_child(node, node->get_size() - 1);
        //这个更快
        parent->set_key(1, neighbor_node->get_key(0));
    }
    else{
        int size = neighbor_node->get_size();
        node->insert_pair(0, neighbor_node->get_key(size-1), *(neighbor_node->get_rid(size-1)));
        neighbor_node->erase_pair(size-1);
        maintain_child(node, 0);
        parent->set_key(index, node->get_key(0));
    }
    // 2. 从neighbor_node中移动一个键值对到node结点中
    // 3. 更新父节点中的相关信息，并且修改移动键值对对应孩字结点的父结点信息（maintain_child函数）
    // 注意：neighbor_node的位置不同，需要移动的键值对不同，需要分类讨论
}

/**
 * @brief 合并(Coalesce)函数是将node和其直接前驱进行合并，也就是和它左边的neighbor_node进行合并；
 * 假设node一定在右边。如果上层传入的index=0，说明node在左边，那么交换node和neighbor_node，保证node在右边；合并到左结点，实际上就是删除了右结点；
 * Move all the key & value pairs from one page to its sibling page, and notify buffer pool manager to delete this page.
 * Parent page must be adjusted to take info of deletion into account. Remember to deal with coalesce or redistribute
 * recursively if necessary.
 *
 * @param neighbor_node sibling page of input "node" (neighbor_node是node的前结点)
 * @param node input from method coalesceOrRedistribute() (node结点是需要被删除的)
 * @param parent parent page of input "node"
 * @param index node在parent中的rid_idx
 * @return true means parent node should be deleted, false means no deletion happend
 * @note Assume that *neighbor_node is the left sibling of *node (neighbor -> node)
 */
bool IxIndexHandle::coalesce(IxNodeHandle *neighbor_node, IxNodeHandle *node, IxNodeHandle *parent, int index,
                             Transaction *transaction, bool *root_is_latched) {
    // Todo:
    // 1. 用index判断neighbor_node是否为node的前驱结点，若不是则交换两个结点，让neighbor_node作为左结点，node作为右结点
    if(index == 0){
        std::swap(neighbor_node, node);
        index = 1;
    }
    // 2. 把node结点的键值对移动到neighbor_node中，并更新node结点孩子结点的父节点信息（调用maintain_child函数）
    neighbor_node->insert_pairs(neighbor_node->get_size(), node->get_key(0), node->get_rid(0), node->get_size());
    if (!neighbor_node->is_leaf_page()){
        for(int i = neighbor_node->get_size() - node->get_size();i < neighbor_node->get_size(); i++){
            maintain_child(neighbor_node, i);
        }
    }
    // 提示：如果是叶子结点且为最右叶子结点，需要更新file_hdr_.last_leaf
    else if(file_hdr_->last_leaf_ == node->get_page_no()){
        file_hdr_->last_leaf_ = node->get_prev_leaf();
    }
    // 3. 释放和删除node结点，并删除parent中node结点的信息，返回parent是否需要被删除
    if(node->is_leaf_page()){
        erase_leaf(node);
    }
    int origin_key_num = parent->get_size();
    int afer_key_num = parent->remove(parent->get_key(index));
    release_node_handle(*node);
    if(origin_key_num == afer_key_num){
        return false;
    }
    
    return coalesce_or_redistribute(parent, nullptr, root_is_latched);
}

/**
 * @brief 这里把iid转换成了rid，即iid的slot_no作为node的rid_idx(key_idx)
 * node其实就是把slot_no作为键值对数组的下标
 * 换而言之，每个iid对应的索引槽存了一对(key,rid)，指向了(要建立索引的属性首地址,插入/删除记录的位置)
 *
 * @param iid
 * @return Rid
 * @note iid和rid存的不是一个东西，rid是上层传过来的记录位置，iid是索引内部生成的索引槽位置
 */
Rid IxIndexHandle::get_rid(const Iid &iid) const {
    IxNodeHandle *node = fetch_node(iid.page_no);
    if (iid.slot_no >= node->get_size()) {
        throw IndexEntryNotFoundError();
    }
    Rid rid = *node->get_rid(iid.slot_no);
    buffer_pool_manager_->unpin_page(node->get_page_id(), false);  // unpin it!
    delete node;
    return rid;
}

/**
 * @brief FindLeafPage + lower_bound
 *
 * @param key
 * @return Iid
 * @note 上层传入的key本来是int类型，通过(const char *)&key进行了转换
 * 可用*(int *)key转换回去
 */
Iid IxIndexHandle::lower_bound(const char *key) {
    std::scoped_lock lock{root_latch_};
    std::pair<IxNodeHandle *, bool> node_pair = find_leaf_page(key, Operation::FIND, nullptr);
    IxNodeHandle *node = node_pair.first;
    //查找键为key的键值对在叶子节点中的序号
    if(!node){
        delete node;
        return Iid{-1, -1};
    }
    int key_num = node->lower_bound(key);
    Iid iid;
    if(key_num == node->get_size() && node->get_page_no() != file_hdr_->last_leaf_){
        //没有找到,
        iid = {.page_no = node->get_next_leaf(), .slot_no = 0};
    }
    else{
        iid = {.page_no = node->get_page_no(), .slot_no = key_num};
    }
    buffer_pool_manager_->unpin_page(node->get_page_id(), false);
    delete node;
	return iid;
    //??????
    //return Iid{-1, -1};
}

/**
 * @brief FindLeafPage + upper_bound
 *
 * @param key
 * @return Iid
 */
Iid IxIndexHandle::upper_bound(const char *key) {
    std::scoped_lock lock{root_latch_};
    std::pair<IxNodeHandle *, bool> node_pair = find_leaf_page(key, Operation::FIND, nullptr);
    IxNodeHandle *node = node_pair.first;
    //查找键为key的键值对在叶子节点中的序号
    if(!node){
        return Iid{-1, -1};
    }
    int key_num = node->upper_bound(key);
    Iid iid;
    if(key_num == node->get_size() && node->get_page_no() != file_hdr_->last_leaf_){
        //没有找到,
        iid = {.page_no = node->get_next_leaf(), .slot_no = 0};
    }
    else{
        iid = {.page_no = node->get_page_no(), .slot_no = key_num};
    }
    buffer_pool_manager_->unpin_page(node->get_page_id(), false);
    delete node;
	return iid;
}

/**
 * @brief 指向最后一个叶子的最后一个结点的后一个
 * 用处在于可以作为IxScan的最后一个
 *
 * @return Iid
 */
Iid IxIndexHandle::leaf_end() const {
    IxNodeHandle *node = fetch_node(file_hdr_->last_leaf_);
    Iid iid = {.page_no = file_hdr_->last_leaf_, .slot_no = node->get_size()};
    buffer_pool_manager_->unpin_page(node->get_page_id(), false);  // unpin it!
    delete node;
    return iid;
}

/**
 * @brief 指向第一个叶子的第一个结点
 * 用处在于可以作为IxScan的第一个
 *
 * @return Iid
 */
Iid IxIndexHandle::leaf_begin() const {
    Iid iid = {.page_no = file_hdr_->first_leaf_, .slot_no = 0};
    return iid;
}

/**
 * @brief 获取一个指定结点
 *
 * @param page_no
 * @return std::shared_ptr<IxNodeHandle> 
 * @note pin the page, remember to unpin it outside!
 */
IxNodeHandle *IxIndexHandle::fetch_node(int page_no) const {
    Page *page = buffer_pool_manager_->fetch_page(PageId{fd_, page_no});
    // std::cout<<"fetch_node:"<<&page<<" "<<fd_<<" "<<page->get_page_id().fd<<" "<<page->get_page_id().page_no<<std::endl;
    // auto node = std::make_shared<IxNodeHandle>(file_hdr_, page); 
    IxNodeHandle *node = new IxNodeHandle(file_hdr_, page);
    return node;
}

/**
 * @brief 创建一个新结点
 *
 * @return std::shared_ptr<IxNodeHandle> 
 * @note pin the page, remember to unpin it outside!
 * 注意：对于Index的处理是，删除某个页面后，认为该被删除的页面是free_page
 * 而first_free_page实际上就是最新被删除的页面，初始为IX_NO_PAGE
 * 在最开始插入时，一直是create node，那么first_page_no一直没变，一直是IX_NO_PAGE
 * 与Record的处理不同，Record将未插入满的记录页认为是free_page
 */
IxNodeHandle *IxIndexHandle::create_node() {
    IxNodeHandle *node;
    file_hdr_->num_pages_++;

    PageId new_page_id = {.fd = fd_, .page_no = INVALID_PAGE_ID};
    // 从3开始分配page_no，第一次分配之后，new_page_id.page_no=3，file_hdr_.num_pages=4
    Page *page = buffer_pool_manager_->new_page(&new_page_id);
    node = new IxNodeHandle(file_hdr_, page);
    // std::cout<<"page_add:"<<fd_<<" "<<"page_no"<<page->get_page_id().page_no<<std::endl;
    return node;
}

/**
 * @brief 从node开始更新其父节点的第一个key，一直向上更新直到根节点
 *
 * @param node
 */
void IxIndexHandle::maintain_parent(IxNodeHandle *node) {
    IxNodeHandle *curr = node;
    while (curr->get_parent_page_no() != IX_NO_PAGE) {
        // Load its parent
        IxNodeHandle *parent = fetch_node(curr->get_parent_page_no());
        int rank = parent->find_child(curr);
        char *parent_key = parent->get_key(rank);
        char *child_first_key = curr->get_key(0);
        if (memcmp(parent_key, child_first_key, file_hdr_->col_tot_len_) == 0) {
            assert(buffer_pool_manager_->unpin_page(parent->get_page_id(), true));
            delete parent;
            break;
        }
        memcpy(parent_key, child_first_key, file_hdr_->col_tot_len_);  // 修改了parent node
        curr = parent;

        assert(buffer_pool_manager_->unpin_page(parent->get_page_id(), true));
        delete parent;
    }
}

/**
 * @brief 要删除leaf之前调用此函数，更新leaf前驱结点的next指针和后继结点的prev指针
 *
 * @param leaf 要删除的leaf
 */
void IxIndexHandle::erase_leaf(IxNodeHandle *leaf) {
    assert(leaf->is_leaf_page());

    IxNodeHandle *prev = fetch_node(leaf->get_prev_leaf());
    prev->set_next_leaf(leaf->get_next_leaf());
    buffer_pool_manager_->unpin_page(prev->get_page_id(), true);
    delete prev;

    IxNodeHandle *next = fetch_node(leaf->get_next_leaf());
    next->set_prev_leaf(leaf->get_prev_leaf());  // 注意此处是SetPrevLeaf()
    buffer_pool_manager_->unpin_page(next->get_page_id(), true);
    delete next;
}

/**
 * @brief 删除node时，更新file_hdr_.num_pages
 *
 * @param node
 */
void IxIndexHandle::release_node_handle(IxNodeHandle &node) {
    file_hdr_->num_pages_--;
}

/**
 * @brief 将node的第child_idx个孩子结点的父节点置为node
 */
void IxIndexHandle::maintain_child(IxNodeHandle *node, int child_idx) {
    if (!node->is_leaf_page()) {
        //  Current node is inner node, load its child and set its parent to current node
        int child_page_no = node->value_at(child_idx);
        IxNodeHandle *child = fetch_node(child_page_no);
        child->set_parent_page_no(node->get_page_no());
        buffer_pool_manager_->unpin_page(child->get_page_id(), true);
        delete child;
    }
}