/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "planner.h"

#include <memory>

#include "execution/executor_delete.h"
#include "execution/executor_index_scan.h"
#include "execution/executor_insert.h"
#include "execution/executor_nestedloop_join.h"
#include "execution/executor_projection.h"
#include "execution/executor_seq_scan.h"
#include "execution/executor_update.h"
#include "index/ix.h"
#include "record_printer.h"

#include <algorithm>    // 用于自定义函数最左匹配
// 为了load解析csv文件添加
#include <fstream>
#include <sstream>
// 目前的索引匹配规则为：完全匹配索引字段，且全部为单点查询，不会自动调整where条件的顺序
// bool Planner::get_index_cols(std::string tab_name, std::vector<Condition> curr_conds, std::vector<std::string>& index_col_names) {
//     index_col_names.clear();
//     for(auto& cond: curr_conds) {
//         if(cond.is_rhs_val && cond.op == OP_EQ && cond.lhs_col.tab_name.compare(tab_name) == 0)
//             index_col_names.push_back(cond.lhs_col.col_name);
//     }
//     TabMeta& tab = sm_manager_->db_.get_table(tab_name);
//     if(tab.is_index(index_col_names)) return true;
//     return false;
// }
/*
提示：
在大赛提供的框架中，只有查询条件与索引完全一致，并且是单点查询时，才使用索引来进行查询，你需要对索引匹配规则进行修改，要求使用最左匹配原则。

最左匹配原则：最左优先，以最左边的为起点任何连续的索引都能匹配上。同时遇到范围查询(>、<、between、like)就会停止匹配。
// node:目前只支持右侧是值，不是tbl.col这种的where找索引
0.查看tab中索引数量为0直接返回false没有索引
1.首先改写where语句中使用=的部分，选取条件左侧和要找的表名相同的条件，它们作为查找的索引列的最左侧
    1.1将所有=条件的列拿出来 equal_cols
    1.2后边的不是等于的（>=这种也不是=）列都拿出来  not_equal_cols
2.通过比较tab中的索引是否存在equal_cols的所有排列组合（索引和排列的顺序必须一致）
    2.1equal_cols不为空，逐个查看tab中索引，根据索引中的列个数n排列equal_cols的前n列或所有列（可能相等的列个数<n）不同的组合，查看是否有匹配的
    2.2not_equal_cols不为空，拿出一列看是否能匹配2.1的索引的后一列
*/
// 判断equal_cols的一个排列是否能与任一索引的cols按最左原则匹配
// 返回的<int, vector<int>>分别为匹配了几列，匹配的索引为tab.indexes[?]的?,没有匹配返回<0, 空>
std::pair<int, int> check_prefix_match(const std::vector<std::string>& perm, const std::vector<IndexMeta>& indexes, int max_index_cols_num) {
    // permutation排列：为一种where条件包含的equal_cols所有列的顺序
    int ret_cols_num = 0;
    int ret_index_id = -1;
    int index_id = -1;
    for (const auto& index : indexes) {
        index_id += 1;
        size_t match_cols_count = 0;
        for (size_t i = 0; i < std::min(perm.size(), index.cols.size()); ++i) {
            if (perm[i] == index.cols[i].name) {
                ++match_cols_count;
            } else {
                break; // 不匹配，跳出循环
            }
        }

        // 到这里一定是匹配了某个索引
        // 如果当前等于部分条件的排列与所有索引已经有最大匹配，则返回
        if (match_cols_count == (size_t)(max_index_cols_num)) {
            ret_cols_num = match_cols_count;
            ret_index_id = index_id;
            return {ret_cols_num, ret_index_id};
        }
        else if (match_cols_count > ret_cols_num){
            ret_cols_num = match_cols_count;
            ret_index_id = index_id;
        }
    }
    if(ret_cols_num != 0)
    {   // 如果匹配成功过索引
        return  {ret_cols_num, ret_index_id};
    }
    else
    {   // 没匹配成功过索引
        return {0, -1};
    }
}

/*
返回情况:
1.无索引:false
2.无条件,有索引:true tab.indexes[0]
3.没有不相等的条件列，相等的条件列又没有匹配任何索引，返回false
4.没有不相等的条件，并且匹配了相等的条件，直接挑一个索引赋值index_col_names，返回true即可
5.索引只能匹配到相等的条件列，不是相等的条件列不能再匹配（因为索引没这么长）

*/
bool Planner::get_index_cols(std::string tab_name, std::vector<Condition> curr_conds, std::vector<std::string>& index_col_names) {
    index_col_names.clear();
    std::vector<std::string> equal_cols;
    std::vector<std::string> not_equal_cols;

    TabMeta& tab = sm_manager_->db_.get_table(tab_name);
    size_t max_index_cols_num = 0;  // 所有存在的索引中，列数量的最大值，例如{id},{id,name}，应为2
    if (tab.indexes.size() == 0)    // 没有索引，返回false
    {
        return false;
    }
    else    // 否则找到max_index_cols_num
    {
        for (const auto& index : tab.indexes)
        {
            size_t index_cols_num = index.col_num;
            if (index_cols_num > max_index_cols_num)
            {
                max_index_cols_num = index_cols_num;
            }
        }
    }

    // 而且如果没有条件应该直接返回吗？
    // 这里下边的加入列，还应该防止重复 stu.id = 1 and stu.id > 1是错误的, stu.id = 1 and stu.id > 0怎么处理？
    // op = OP_INVALID的时候还没处理lun
    for(auto& cond: curr_conds) {
        if(cond.is_rhs_val && cond.op == OP_EQ && cond.lhs_col.tab_name.compare(tab_name) == 0){
            if (std::find(equal_cols.begin(), equal_cols.end(), cond.lhs_col.col_name) == equal_cols.end()) 
            {   
                // 避免索引列重复，如果不重复就push_back
                equal_cols.push_back(cond.lhs_col.col_name);
            }
        }
        else if(cond.is_rhs_val && cond.op != OP_EQ && cond.lhs_col.tab_name.compare(tab_name) == 0){
            if (std::find(not_equal_cols.begin(), not_equal_cols.end(), cond.lhs_col.col_name) == not_equal_cols.end())
            {   
                not_equal_cols.push_back(cond.lhs_col.col_name);
            }
        }
        else if(!cond.is_rhs_val && cond.op == OP_EQ && (cond.lhs_col.tab_name.compare(tab_name) == 0)){
            if (std::find(equal_cols.begin(), equal_cols.end(), cond.lhs_col.col_name) == equal_cols.end())
            {   
                equal_cols.push_back(cond.lhs_col.col_name);    // where t.id = d.idd     tab_name=t,push_back(id)
            }
        }
        else if(!cond.is_rhs_val && cond.op == OP_EQ && (cond.rhs_col.tab_name.compare(tab_name) == 0)){
            if (std::find(equal_cols.begin(), equal_cols.end(), cond.lhs_col.col_name) == equal_cols.end())
            {   
                equal_cols.push_back(cond.rhs_col.col_name);    // where t.id = d.idd     tab_name=d,push_back(idd)
            }
        }
    }

    // 没有where条件的时候如果有索引也要走索引
    if ((equal_cols.size() == 0) && (not_equal_cols.size() == 0))
    {   
        // 全表扫描，所有索引效率都一样，直接找这个表中随便的一个索引，indexes[0]肯定存在
        for(auto col:tab.indexes[0].cols){
            index_col_names.push_back(col.name);
        }
        return true;
    }

    // 对equal_cols排序，确保生成所有可能的排列
    int match_cols_num = 0; // match_cols_num<=max_index_cols_num
    int match_index_id = -1;
    // 如果不是找到的完美匹配，找到的可能有多个索引都符合，全都记录下来，它们匹配的列数都是match_cols_num
    std::vector<int> match_equal_index_id; 

    std::unordered_map<std::string, bool> col_exists;
    for(auto equal_col:equal_cols){
        col_exists[equal_col] = true;
    }
    if (equal_cols.size() > 0)
    {   
        int index_id = -1;
        for(auto index:tab.indexes){
            int temp_match_cols_num = 0;
            index_id += 1;
            for(auto index_col: index.cols){
                if(col_exists[index_col.name] == true){
                    // 从左开始匹配，意味着左边能匹配的最多个数
                    temp_match_cols_num += 1;
                }else{
                    if(temp_match_cols_num != 0){
                        if ((temp_match_cols_num == match_cols_num) && (match_cols_num != 0))
                        {
                            match_equal_index_id.push_back(index_id);
                        }
                        else if (temp_match_cols_num > match_cols_num)
                        {   
                            match_equal_index_id.clear();
                            match_cols_num = temp_match_cols_num;
                            match_index_id = index_id;
                            match_equal_index_id.push_back(index_id);
                        }
                    }
                    else{
                        // do nothing
                    }// if
                }// if
            }// for
        }
    }
    
    if (not_equal_cols.size() == 0)
    {   
        if (match_cols_num == 0)
        {
            return false;   // 如果没有不相等的条件列，相等的条件列又没有匹配任何索引，返回false
        }
        else
        {   
            // match_cols_num > 0的情况
            // 如果没有不相等的条件，并且匹配了相等的条件，直接挑一个索引赋值index_col_names，返回true即可
            for (int i = 0; i < tab.indexes[match_index_id].cols.size(); ++i) {
                index_col_names.push_back(tab.indexes[match_index_id].cols[i].name);
            }
            return true;
        }
    }
    else
    {
        // 如果有not_equal_cols
        if ((match_cols_num == max_index_cols_num) && (max_index_cols_num != 0))
        {
            // 索引只能匹配到相等的条件列，不是相等的条件列不能再匹配（因为索引没这么长）
            for (int i = 0; i < tab.indexes[match_index_id].cols.size(); ++i) {
                index_col_names.push_back(tab.indexes[match_index_id].cols[i].name);
            }
            return true;
        }
    }
    

    // 这里为进行不相等列匹配
    // 可能匹配的索引还能匹配下一列，比较不相等条件是否可以匹配一个之后的索引列
    if (match_cols_num == 0)
    {   // 表示没有相等的where条件或者相等条件列的都没匹配到
        bool have_matched = false;
        for (const auto& index : tab.indexes)
        {   
            have_matched = false;
            for (const auto& col : not_equal_cols)
            {
                if (col == index.cols[0].name)
                {
                    for (int i = 0; i < index.cols.size(); ++i) {
                        index_col_names.push_back(index.cols[i].name);
                    }
                    have_matched =true;
                    break;
                }
            }
            if(have_matched) {
                return true;
            }
        }
        if(have_matched == false) {
            return false;
        }
    }
    else
    {   // 继续往后匹配，但要确保vector没出界
        bool have_matched = false;
        for (const auto& index_id : match_equal_index_id)
        {   
            have_matched = false;
            if (match_cols_num == tab.indexes[index_id].cols.size())
            {   
                continue;   // 这个索引没有下一个能匹配的不是相等条件的列了
            }
            else
            {
                for (const auto& col : not_equal_cols)
                {   
                    if (col == tab.indexes[index_id].cols[match_cols_num].name)
                    {   
                        index_col_names.clear();
                        for (int i = 0; i < tab.indexes[index_id].cols.size(); ++i) {
                            index_col_names.push_back(tab.indexes[index_id].cols[i].name);
                        }
                        have_matched =true;
                        break;
                    }
                }
            }
            
            if(have_matched)
            {
                return true;
            }
        }
        // 到这里说明match_cols_num!=0,但索引都没匹配上后一列，随便找个匹配的索引返回就行了
        if(have_matched == false)
        {   
            index_col_names.clear();
            for (int i = 0; i < tab.indexes[match_index_id].cols.size(); ++i) {
                index_col_names.push_back(tab.indexes[match_index_id].cols[i].name);
            }
            return true;
        }
    }

}

/**
 * @brief 表算子条件谓词生成
 *
 * @param conds 条件
 * @param tab_names 表名
 * @return std::vector<Condition>
 */
std::vector<Condition> pop_conds(std::vector<Condition> &conds, std::string tab_names) {
    // auto has_tab = [&](const std::string &tab_name) {
    //     return std::find(tab_names.begin(), tab_names.end(), tab_name) != tab_names.end();
    // };
    std::vector<Condition> solved_conds;
    auto it = conds.begin();
    while (it != conds.end()) {
        if ((tab_names.compare(it->lhs_col.tab_name) == 0 && it->is_rhs_val) || (it->lhs_col.tab_name.compare(it->rhs_col.tab_name) == 0)) {
            solved_conds.emplace_back(std::move(*it));
            it = conds.erase(it);
        } 
        else{
            it++;
        }
    }
    return solved_conds;
}

std::vector<Condition> pop_conds_for_scan(std::vector<Condition> &conds, std::string tab_names) {
    // auto has_tab = [&](const std::string &tab_name) {
    //     return std::find(tab_names.begin(), tab_names.end(), tab_name) != tab_names.end();
    // };
    std::vector<Condition> solved_conds;
    auto it = conds.begin();
    while (it != conds.end()) {
        if ((tab_names.compare(it->lhs_col.tab_name) == 0 && it->is_rhs_val) || (it->lhs_col.tab_name.compare(it->rhs_col.tab_name) == 0)) {
            solved_conds.emplace_back(std::move(*it));
            it = conds.erase(it);
        }
        else{
            it++;
        }
    }
    return solved_conds;
}

int push_conds(Condition *cond, std::shared_ptr<Plan> plan)
{
    if(auto x = std::dynamic_pointer_cast<ScanPlan>(plan))
    {
        if(x->tab_name_.compare(cond->lhs_col.tab_name) == 0) {
            return 1;
        } else if(x->tab_name_.compare(cond->rhs_col.tab_name) == 0){
            return 2;
        } else {
            return 0;
        }
    }
    else if(auto x = std::dynamic_pointer_cast<JoinPlan>(plan))
    {
        int left_res = push_conds(cond, x->left_);
        // 条件已经下推到左子节点
        if(left_res == 3){
            return 3;
        }
        int right_res = push_conds(cond, x->right_);
        // 条件已经下推到右子节点
        if(right_res == 3){
            return 3;
        }
        // 左子节点或右子节点有一个没有匹配到条件的列
        if(left_res == 0 || right_res == 0) {
            return left_res + right_res;
        }
        // 左子节点匹配到条件的右边
        if(left_res == 2) {
            // 需要将左右两边的条件变换位置
            std::map<CompOp, CompOp> swap_op = {
                {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
            };
            std::swap(cond->lhs_col, cond->rhs_col);
            cond->op = swap_op.at(cond->op);
        }
        x->conds_.emplace_back(std::move(*cond));
        return 3;
    }
    return false;
}

std::shared_ptr<Plan> pop_scan(int *scantbl, std::string table, std::vector<std::string> &joined_tables, 
                std::vector<std::shared_ptr<Plan>> plans)
{
    for (size_t i = 0; i < plans.size(); i++) {
        auto x = std::dynamic_pointer_cast<ScanPlan>(plans[i]);
        if(x->tab_name_.compare(table) == 0)
        {
            scantbl[i] = 1;
            joined_tables.emplace_back(x->tab_name_);
            return plans[i];
        }
    }
    return nullptr;
}


std::shared_ptr<Query> Planner::logical_optimization(std::shared_ptr<Query> query, Context *context)
{
    
    //TODO 实现逻辑优化规则

    return query;
}

std::shared_ptr<Plan> Planner::physical_optimization(std::shared_ptr<Query> query, Context *context)
{
    std::shared_ptr<Plan> plan = make_one_rel(query);
    //增加聚合算子
    std::shared_ptr<Plan> new_plan=generate_agg_plan(query,std::move(plan));
    
    // 其他物理优化

    // 处理orderby
    plan = generate_sort_plan(query, std::move(new_plan)); 

    return plan;
}



std::shared_ptr<Plan> Planner::make_one_rel(std::shared_ptr<Query> query)
{
    auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse);
    std::vector<std::string> tables = query->tables;
    // // Scan table , 生成表算子列表tab_nodes
    std::vector<std::shared_ptr<Plan>> table_scan_executors(tables.size());
    for (size_t i = 0; i < tables.size(); i++) {
        auto curr_conds = pop_conds(query->conds, tables[i]);
        // int index_no = get_indexNo(tables[i], curr_conds);
        std::vector<std::string> index_col_names;
        bool index_exist = get_index_cols(tables[i], curr_conds, index_col_names);
        auto curr_cond_for_scan_=pop_conds_for_scan(curr_conds,tables[i]);
        if (index_exist == false) {  // 该表没有索引
            index_col_names.clear();
            table_scan_executors[i] = 
                std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, tables[i], curr_cond_for_scan_, index_col_names);
        } else {  // 存在索引
            table_scan_executors[i] =
                std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, tables[i], curr_cond_for_scan_, index_col_names);
        }
    }
    // 只有一个表，不需要join。
    if(tables.size() == 1)
    {
        return table_scan_executors[0];
    }
    // 获取where条件
    auto conds = std::move(query->conds);
    std::shared_ptr<Plan> table_join_executors;
    
    int scantbl[tables.size()];
    for(size_t i = 0; i < tables.size(); i++)
    {
        scantbl[i] = -1;
    }
    // 假设在ast中已经添加了jointree，这里需要修改的逻辑是，先处理jointree，然后再考虑剩下的部分
    if(conds.size() >= 1)
    {
        // 有连接条件

        // 根据连接条件，生成第一层join
        std::vector<std::string> joined_tables(tables.size());
        auto it = conds.begin();
        while (it != conds.end()) {
            std::shared_ptr<Plan> left , right;
            left = pop_scan(scantbl, it->lhs_col.tab_name, joined_tables, table_scan_executors);
            right = pop_scan(scantbl, it->rhs_col.tab_name, joined_tables, table_scan_executors);
            std::vector<Condition> join_conds{*it};
            //建立join
            // 判断使用哪种join方式
            if(enable_nestedloop_join && enable_sortmerge_join) {
                // 默认nested loop join
                table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(left), std::move(right), join_conds, tables);
            } else if(enable_nestedloop_join) {
                table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(left), std::move(right), join_conds, tables);
            } else if(enable_sortmerge_join) {
                table_join_executors = std::make_shared<JoinPlan>(T_SortMerge, std::move(left), std::move(right), join_conds, tables);
            } else {
                // error
                throw RMDBError("No join executor selected!");
            }

            // table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(left), std::move(right), join_conds);
            it = conds.erase(it);
            break;
        }
        // 根据连接条件，生成第2-n层join
        it = conds.begin();
        while (it != conds.end()) {
            std::shared_ptr<Plan> left_need_to_join_executors = nullptr;
            std::shared_ptr<Plan> right_need_to_join_executors = nullptr;
            bool isneedreverse = false;
            if (std::find(joined_tables.begin(), joined_tables.end(), it->lhs_col.tab_name) == joined_tables.end()) {
                left_need_to_join_executors = pop_scan(scantbl, it->lhs_col.tab_name, joined_tables, table_scan_executors);
            }
            if (std::find(joined_tables.begin(), joined_tables.end(), it->rhs_col.tab_name) == joined_tables.end()) {
                right_need_to_join_executors = pop_scan(scantbl, it->rhs_col.tab_name, joined_tables, table_scan_executors);
                isneedreverse = true;
            } 

            if(left_need_to_join_executors != nullptr && right_need_to_join_executors != nullptr) {
                std::vector<Condition> join_conds{*it};
                std::shared_ptr<Plan> temp_join_executors = std::make_shared<JoinPlan>(T_NestLoop, 
                                                                    std::move(left_need_to_join_executors), 
                                                                    std::move(right_need_to_join_executors), 
                                                                    join_conds,
                                                                    tables);
                table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(temp_join_executors), 
                                                                    std::move(table_join_executors), 
                                                                    std::vector<Condition>(),
                                                                    tables);
            } else if(left_need_to_join_executors != nullptr || right_need_to_join_executors != nullptr) {
                if(isneedreverse) {
                    std::map<CompOp, CompOp> swap_op = {
                        {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
                    };
                    std::swap(it->lhs_col, it->rhs_col);
                    it->op = swap_op.at(it->op);
                    left_need_to_join_executors = std::move(right_need_to_join_executors);
                }
                std::vector<Condition> join_conds{*it};
                table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(left_need_to_join_executors), 
                                                                    std::move(table_join_executors), join_conds, tables);
            } else {
                push_conds(std::move(&(*it)), table_join_executors);
            }
            it = conds.erase(it);
        }
    } else {
        table_join_executors = table_scan_executors[0];
        scantbl[0] = 1;
    }

    //连接剩余表
    for (size_t i = 0; i < tables.size(); i++) {
        if(scantbl[i] == -1) {
            table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(table_scan_executors[i]), 
                                                    std::move(table_join_executors), std::vector<Condition>(), tables);
        }
    }

    return table_join_executors;

}

//增加聚合函数逻辑计划生成
std::shared_ptr<Plan> Planner::generate_agg_plan(std::shared_ptr<Query> query, std::shared_ptr<Plan> plan){
    auto x=std::dynamic_pointer_cast<ast::SelectStmt>(query->parse);
    if(x->aggregatestmts.size()==0){
        return plan;
    }
    std::vector<std::string> tables = query->tables;
    std::vector<ColMeta> all_cols;
    for (auto &sel_tab_name : tables) {
        // 这里db_不能写成get_db(), 注意要传指针
        const auto &sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
        all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
    }
    std::vector<AggregateFunction> aggregatefunctions;
    std::vector<TabCol> groupby_cols;
    //记录聚合函数需要输出的列
    std::vector<Agg_Cols> agg_cols;
    // std::vector<TabCol> out_agg_cols;
    //??对query处理
    query->out_agg_cols.clear();
    for(auto agg_col:x->agg_cols){
        if (auto aggcol = std::dynamic_pointer_cast<ast::Col>(agg_col)){
            TabCol sel_col;
            
            for(auto &col : all_cols){
                if(col.name.compare(aggcol->col_name) == 0 ){
                    sel_col = {.tab_name = col.tab_name, .col_name = col.name};
                    break;
                }
            }
            agg_cols.emplace_back(false,sel_col,ast::AggregateCOUNT,"");
            groupby_cols.push_back(sel_col);
            query->out_agg_cols.push_back(sel_col);
        }else if(auto aggregate = std::dynamic_pointer_cast<ast::AggregateStmt>(agg_col)){
            TabCol sel_col;
            TabCol out_col;
            if(aggregate->col==nullptr){
                sel_col = {.tab_name ="", .col_name ="*"};
                out_col={.tab_name ="", .col_name = aggregate->as_col};
            }else{
                for(auto &col : all_cols){
                    if(col.name.compare(aggregate->col->col_name) == 0 ){
                        sel_col = {.tab_name = col.tab_name, .col_name = col.name};
                        out_col={.tab_name =col.tab_name, .col_name = aggregate->as_col};
                        break;
                    }
                }
            }
            agg_cols.emplace_back(true,sel_col,aggregate->type,aggregate->as_col);
            aggregatefunctions.emplace_back(aggregate->type,aggregate->as_col,sel_col);
            query->out_agg_cols.push_back(out_col);
        }
    }
    
    return std::make_shared<AggPlan>(T_Agg,aggregatefunctions,groupby_cols,plan,agg_cols,query->out_agg_cols,query->having);
}

std::shared_ptr<Plan> Planner::generate_sort_plan(std::shared_ptr<Query> query, std::shared_ptr<Plan> plan)
{
    auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse);
    if(!x->has_sort) {
        return plan;
    }
    std::vector<std::string> tables = query->tables;
    std::vector<ColMeta> all_cols;
    for (auto &sel_tab_name : tables) {
        // 这里db_不能写成get_db(), 注意要传指针
        const auto &sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
        all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
    }
    TabCol sel_col;
    for (auto &col : all_cols) {
        if(col.name.compare(x->order->cols->col_name) == 0 )
        sel_col = {.tab_name = col.tab_name, .col_name = col.name};
    }
    return std::make_shared<SortPlan>(T_Sort, std::move(plan), sel_col, 
                                    x->order->orderby_dir == ast::OrderBy_DESC);
}


/**
 * @brief select plan 生成
 *
 * @param sel_cols select plan 选取的列
 * @param tab_names select plan 目标的表
 * @param conds select plan 选取条件
 */
std::shared_ptr<Plan> Planner::generate_select_plan(std::shared_ptr<Query> query, Context *context) {
    //逻辑优化
    query = logical_optimization(std::move(query), context);

    //物理优化
    auto sel_cols = query->cols;
    std::shared_ptr<Plan> plannerRoot = physical_optimization(query, context);
    plannerRoot = std::make_shared<ProjectionPlan>(T_Projection, std::move(plannerRoot), 
                                                        std::move(sel_cols),query->out_agg_cols);

    return plannerRoot;
}

// 生成DDL语句和DML语句的查询执行计划
std::shared_ptr<Plan> Planner::do_planner(std::shared_ptr<Query> query, Context *context)
{
    std::shared_ptr<Plan> plannerRoot;
    if (auto x = std::dynamic_pointer_cast<ast::CreateTable>(query->parse)) {
        // create table;
        std::vector<ColDef> col_defs;
        for (auto &field : x->fields) {
            if (auto sv_col_def = std::dynamic_pointer_cast<ast::ColDef>(field)) {
                ColDef col_def = {.name = sv_col_def->col_name,
                                  .type = interp_sv_type(sv_col_def->type_len->type),
                                  .len = sv_col_def->type_len->len};
                col_defs.push_back(col_def);
            } else {
                throw InternalError("Unexpected field type");
            }
        }
        plannerRoot = std::make_shared<DDLPlan>(T_CreateTable, x->tab_name, std::vector<std::string>(), col_defs);
    } else if (auto x = std::dynamic_pointer_cast<ast::DropTable>(query->parse)) {
        // drop table;
        plannerRoot = std::make_shared<DDLPlan>(T_DropTable, x->tab_name, std::vector<std::string>(), std::vector<ColDef>());
    } else if (auto x = std::dynamic_pointer_cast<ast::CreateIndex>(query->parse)) {
        // create index;
        plannerRoot = std::make_shared<DDLPlan>(T_CreateIndex, x->tab_name, x->col_names, std::vector<ColDef>());
    } else if (auto x = std::dynamic_pointer_cast<ast::DropIndex>(query->parse)) {
        // drop index
        plannerRoot = std::make_shared<DDLPlan>(T_DropIndex, x->tab_name, x->col_names, std::vector<ColDef>());
    } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(query->parse)) {
        // insert;
        plannerRoot = std::make_shared<DMLPlan>(T_Insert, std::shared_ptr<Plan>(),  x->tab_name,  
                                                    query->values, std::vector<Condition>(), std::vector<SetClause>());
    } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(query->parse)) {
        // delete;
        // 生成表扫描方式
        std::shared_ptr<Plan> table_scan_executors;
        // 只有一张表，不需要进行物理优化了
        // int index_no = get_indexNo(x->tab_name, query->conds);
        std::vector<std::string> index_col_names;
        bool index_exist = get_index_cols(x->tab_name, query->conds, index_col_names);
        
        if (index_exist == false) {  // 该表没有索引
            index_col_names.clear();
            table_scan_executors = 
                std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, x->tab_name, query->conds, index_col_names);
        } else {  // 存在索引
            table_scan_executors =
                std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, x->tab_name, query->conds, index_col_names);
        }

        plannerRoot = std::make_shared<DMLPlan>(T_Delete, table_scan_executors, x->tab_name,  
                                                std::vector<Value>(), query->conds, std::vector<SetClause>());
    } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(query->parse)) {
        // update;
        // 生成表扫描方式
        std::shared_ptr<Plan> table_scan_executors;
        // 只有一张表，不需要进行物理优化了
        // int index_no = get_indexNo(x->tab_name, query->conds);
        std::vector<std::string> index_col_names;
        bool index_exist = get_index_cols(x->tab_name, query->conds, index_col_names);

        if (index_exist == false) {  // 该表没有索引
        index_col_names.clear();
            table_scan_executors = 
                std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, x->tab_name, query->conds, index_col_names);
        } else {  // 存在索引
            table_scan_executors =
                std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, x->tab_name, query->conds, index_col_names);
        }
        plannerRoot = std::make_shared<DMLPlan>(T_Update, table_scan_executors, x->tab_name,
                                                     std::vector<Value>(), query->conds, 
                                                     query->set_clauses);
    } else if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse)) {

        std::shared_ptr<plannerInfo> root = std::make_shared<plannerInfo>(x);
        // 生成select语句的查询执行计划
        std::shared_ptr<Plan> projection = generate_select_plan(std::move(query), context);
        plannerRoot = std::make_shared<DMLPlan>(T_select, projection, std::string(), std::vector<Value>(),
                                                    std::vector<Condition>(), std::vector<SetClause>());
    } else {
        throw InternalError("Unexpected AST root");
    }
    return plannerRoot;
}