/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "rm_file_handle.h"
// "rm_file_handle.h"->"rm_defs.h"->"storage/buffer_pool_manager.h"->"errors.h"用于抛出异常


//不加锁的get record
std::unique_ptr<RmRecord> RmFileHandle::get_record_without_lock(const Rid &rid, Context *context) const{
    RmPageHandle page_handle = fetch_page_handle(rid.page_no);
    if(Bitmap::is_set(page_handle.bitmap, rid.slot_no))
    {
        // 存在记录返回对应rid的RmRecord指针
        auto record_ptr(std::make_unique<RmRecord>(file_hdr_.record_size, page_handle.get_slot(rid.slot_no)));
        // 等于下边的语句
        // std::unique_ptr<RmRecord> record_ptr = std::make_unique<RmRecord>();
        // record_ptr->size = page_handle.file_hdr->record_size;
        // record_ptr->data = page_handle.get_slot(rid.slot_no);
        buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), false);
        return record_ptr;
    }
    else
    {   
        // 不存在记录抛出错误
        buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), false);
        throw RecordNotFoundError(rid.page_no, rid.slot_no);
    }
}
/**
 * @description: 获取当前表中记录号为rid的记录
 * @param {Rid&} rid 记录号，指定记录的位置
 * @param {Context*} context
 * @return {unique_ptr<RmRecord>} rid对应的记录对象指针
 */
std::unique_ptr<RmRecord> RmFileHandle::get_record(const Rid& rid, Context* context) const {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 初始化一个指向RmRecord的指针（赋值其内部的data和size）
    if(context!=nullptr){
        // context->lock_mgr_->lock_shared_on_record(context->txn_, rid, fd_);
    }
    return get_record_without_lock(rid,context);
}

/**
 * @description: 在当前表中插入一条记录，不指定插入位置
 * @param {char*} buf 要插入的记录的数据
 * @param {Context*} context
 * @return {Rid} 插入的记录的记录号（位置）
 */
Rid RmFileHandle::insert_record(char* buf, Context* context) {
    // Todo:
    // 1. 获取当前未满的page handle
    // 2. 在page handle中找到空闲slot位置
    // 3. 将buf复制到空闲slot位置
    // 4. 更新page_handle.page_hdr中的数据结构
    // 注意考虑插入一条记录后页面已满的情况，需要更新file_hdr_.first_free_page_no
    RmPageHandle page_handle = create_page_handle();
    int slot_no = Bitmap::first_bit(false, page_handle.bitmap, file_hdr_.num_records_per_page);
    Bitmap::set(page_handle.bitmap, slot_no);
    auto rid = Rid{.page_no = page_handle.page->get_page_id().page_no, .slot_no = slot_no};
    if(context!=nullptr){
        // context->lock_mgr_->lock_exclusive_on_record(context->txn_, rid, fd_,buf);
    }
    memcpy(page_handle.get_slot(slot_no), buf, file_hdr_.record_size);
    // set修改在unpin时写到磁盘
    // Bitmap::set(page_handle.bitmap, slot_no);
    // 更新page_hdr：记录+1，如果页满，通过下一个包含空闲空间页面号赋值first_free_page_no
    page_handle.page_hdr->num_records += 1;
    // std::cout<< "insert record:" << *(int*)buf << " page_no:" << page_handle.page->get_page_id().page_no
    //  << " slot_no:" << slot_no << std::endl;
    // 插入后page中记录已满
    if(page_handle.page_hdr->num_records == file_hdr_.num_records_per_page)
    {
        // 下一空闲页不存在
        if(page_handle.page_hdr->next_free_page_no == INVALID_PAGE_ID)
        {
            file_hdr_.first_free_page_no = INVALID_PAGE_ID;
            disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
        }
        // 下一空闲页存在
        else 
        {   
            file_hdr_.first_free_page_no = page_handle.page_hdr->next_free_page_no;
            page_handle.page_hdr->next_free_page_no = INVALID_PAGE_ID;
            disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
        }
    }
    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
    return Rid{page_handle.page->get_page_id().page_no, slot_no};
}

/**
 * @description: 在当前表中的指定位置插入一条记录
 * @param {Rid&} rid 要插入记录的位置
 * @param {char*} buf 要插入记录的数据
 */
void RmFileHandle::insert_record(const Rid& rid, char* buf) {
    RmPageHandle page_handle = fetch_page_handle(rid.page_no);

    bool cannot_insert = Bitmap::is_set(page_handle.bitmap, rid.slot_no);
    // ??直接覆盖？还是有记录的时候不能覆盖
    // if(cannot_insert)
    // {
    //     throw 
    // }
    if(!cannot_insert)
    {
        memcpy(page_handle.get_slot(rid.slot_no), buf, page_handle.file_hdr->record_size);
    }
    Bitmap::set(page_handle.bitmap, rid.slot_no);
    page_handle.page_hdr->num_records += 1;
    // std::cout<< "insert record(rid):" << *(int*)buf << " page_no:" << rid.page_no
    //  << " slot_no:" << rid.slot_no << std::endl;
    //插入后page中记录已满
    if(page_handle.page_hdr->num_records == file_hdr_.num_records_per_page)
    {   
        // 插入第一个空闲页，不论后边有没有空闲页，都是这个逻辑
        if(file_hdr_.first_free_page_no == rid.page_no)
        {
            file_hdr_.first_free_page_no = page_handle.page_hdr->next_free_page_no;
            disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
            page_handle.page_hdr->next_free_page_no = INVALID_PAGE_ID;
            buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
        }
        // 下一空闲页不存在，插入first->next...的最后一页
        else if(page_handle.page_hdr->next_free_page_no == INVALID_PAGE_ID)
        {   
            // 在最后一页插满记录
            for(int i = file_hdr_.first_free_page_no; /*符合条件直接break*/ ; /*下一个的条件在else里边*/)
            {
                RmPageHandle iter = fetch_page_handle(i);
                if(iter.page_hdr->next_free_page_no == rid.page_no)
                {
                    iter.page_hdr->next_free_page_no = INVALID_PAGE_ID;
                    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
                    break;
                }
                else
                {
                    i = iter.page_hdr->next_free_page_no;
                }
            }
        }
        // 插入到中间空闲页中，插满
        else 
        {   
            for(int i = file_hdr_.first_free_page_no; /*符合条件直接break*/ ; /*下一个的条件在else里边*/)
            {
                RmPageHandle iter = fetch_page_handle(i);
                if(iter.page_hdr->next_free_page_no == rid.page_no)
                {
                    iter.page_hdr->next_free_page_no = page_handle.page_hdr->next_free_page_no;
                    page_handle.page_hdr->next_free_page_no = INVALID_PAGE_ID;
                    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
                    break;
                }
                else
                {
                    i = iter.page_hdr->next_free_page_no;
                }
            }
        }
    }
}

void RmFileHandle::recover_insert_record(const Rid& rid, char* buf) {
    RmPageHandle page_handle = fetch_page_handle(rid.page_no);

    bool cannot_insert = Bitmap::is_set(page_handle.bitmap, rid.slot_no);
    if(!cannot_insert)
    {
        memcpy(page_handle.get_slot(rid.slot_no), buf, page_handle.file_hdr->record_size);
    }
    Bitmap::set(page_handle.bitmap, rid.slot_no);
    page_handle.page_hdr->num_records += 1;
    // std::cout<< "insert record(rid):" << *(int*)buf << " page_no:" << rid.page_no
    //  << " slot_no:" << rid.slot_no << std::endl;
    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
}
/**
 * @description: 删除记录文件中记录号为rid的记录
 * @param {Rid&} rid 要删除的记录的记录号（位置）
 * @param {Context*} context
 */
void RmFileHandle::delete_record(const Rid& rid, Context* context) {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 更新page_handle.page_hdr中的数据结构
    // 注意考虑删除一条记录后页面未满的情况，需要调用release_page_handle()
    if(context!=nullptr){
        // context->lock_mgr_->lock_exclusive_on_record(context->txn_, rid, fd_,nullptr);
    }
    RmPageHandle page_handle = fetch_page_handle(rid.page_no);
    // ??这里空的其实也可以删除，要不要判断
    bool can_delete = Bitmap::is_set(page_handle.bitmap, rid.slot_no);
    if(!can_delete)
    {   
        throw RecordNotFoundError(rid.page_no, rid.slot_no);
    }
    // 不需要修改删除的东西，只需要修改位图为0，这部分就代表无效数据，下次直接在对应位置覆盖即可
    // 即不需要memset(page_handle.get_slot(rid.slot_no), '\0', page_handle.file_hdr->record_size);
    Bitmap::reset(page_handle.bitmap, rid.slot_no);
    page_handle.page_hdr->num_records -= 1;
    if(file_hdr_.num_records_per_page == page_handle.page_hdr->num_records + 1)
    {   
        // 原来此页面记录已满，删除记录后该页面为有空闲记录的页面
        release_page_handle(page_handle);
    }
    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
}
//recover恢复专用delete函数
void RmFileHandle::recover_delete_record(const Rid& rid, Context* context) {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 更新page_handle.page_hdr中的数据结构
    // 注意考虑删除一条记录后页面未满的情况，需要调用release_page_handle()
    RmPageHandle page_handle = fetch_page_handle(rid.page_no);
    // ??这里空的其实也可以删除，要不要判断
    bool can_delete = Bitmap::is_set(page_handle.bitmap, rid.slot_no);
    if(!can_delete)
    {   
        throw RecordNotFoundError(rid.page_no, rid.slot_no);
    }
    // 不需要修改删除的东西，只需要修改位图为0，这部分就代表无效数据，下次直接在对应位置覆盖即可
    // 即不需要memset(page_handle.get_slot(rid.slot_no), '\0', page_handle.file_hdr->record_size);
    Bitmap::reset(page_handle.bitmap, rid.slot_no);
    page_handle.page_hdr->num_records -= 1;
    // if(file_hdr_.num_records_per_page == page_handle.page_hdr->num_records + 1)
    // {   
    //     // 原来此页面记录已满，删除记录后该页面为有空闲记录的页面
    //     release_page_handle(page_handle);
    // }
    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
}

/**
 * @description: 更新记录文件中记录号为rid的记录
 * @param {Rid&} rid 要更新的记录的记录号（位置）
 * @param {char*} buf 新记录的数据
 * @param {Context*} context
 */
void RmFileHandle::update_record(const Rid& rid, char* buf, Context* context) {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 更新记录
    if(context!=nullptr){
        // context->lock_mgr_->lock_exclusive_on_record(context->txn_, rid, fd_,nullptr);
    }
    RmPageHandle page_handle = fetch_page_handle(rid.page_no);
    bool can_update = Bitmap::is_set(page_handle.bitmap, rid.slot_no);
    if(!can_update)
    {   
        throw RecordNotFoundError(rid.page_no, rid.slot_no);
    }
    memcpy(page_handle.get_slot(rid.slot_no), buf, page_handle.file_hdr->record_size);
    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
}

/**
 * 以下函数为辅助函数，仅提供参考，可以选择完成如下函数，也可以删除如下函数，在单元测试中不涉及如下函数接口的直接调用
*/
/**
 * @description: 获取指定页面的页面句柄
 * @param {int} page_no 页面号
 * @return {RmPageHandle} 指定页面的句柄
 */
RmPageHandle RmFileHandle::fetch_page_handle(int page_no) const {
    // Todo:
    // 使用缓冲池获取指定页面，并生成page_handle返回给上层
    // if page_no is invalid, throw PageNotExistError exception
    if(page_no <= INVALID_PAGE_ID || page_no >= file_hdr_.num_pages)
    {
        // 文件存在，则最少有一页，一共file_hdr_.num_pages页，该变量∈[1,...)，
        // page_no从0号页文件头开始，∈[0,file_hdr_.num_pages-1]
        throw PageNotExistError("Error:fetch_page_handle", page_no);
    }
    else
    {
        PageId page_id;
        page_id.fd = fd_;
        page_id.page_no = page_no;
        Page* page = buffer_pool_manager_->fetch_page(page_id);
        return RmPageHandle(&file_hdr_, page);
    }
}

/**
 * @description: 创建一个新的page handle
 * @return {RmPageHandle} 新的PageHandle
 */
// 只会在insert未指定时调用到，如果insert指定到不存在页面插入，直接报错
RmPageHandle RmFileHandle::create_new_page_handle() {
    // Todo:
    // 1.使用缓冲池来创建一个新page
    // 2.更新page handle中的相关信息
    // 3.更新file_hdr_

    // 一个表有没有最大的page页数限制，为8192，在底层有说明
    // ??这里将page_no初始化，new_page会不会更改里边的page_no为新创建的？
    // ??新创建的page里边都是空的吗？还是值有残留？
    PageId newPageId{fd_, file_hdr_.first_free_page_no};
    PageId *pageid = &newPageId;
    Page *page = buffer_pool_manager_->new_page(pageid);
    // 修改file_hdr和page handle
    /*
    RmPageHandle
        const RmFileHdr *file_hdr; 
        // 包括下边
            // int record_size;            
            // int num_pages;              // 文件中分配的页面个数，要改，增加1
            // int num_records_per_page;   
            // int first_free_page_no;     // 要改，改为新创建的这个页号
            // int bitmap_size;            
        Page *page;                 // 新创建page，里边还没有记录，不需要读取page数据
        //对于每⼀个数据⻚，本系统使⽤RmPageHandle类进⾏封装管理，每⼀个数据⻚的开始部分并不直接存放记录数据，⽽是按照以下顺序进⾏存放：
        //| page_lsn_ | page_hdr | bitmap | slots |    
        //所以下边这些都是写page中的初始信息 
        RmPageHdr *page_hdr;
        // RmPageHdr包括：            
            // int next_free_page_no;  // 当前页面满了之后，下一个包含空闲空间的页面号，初始化为-1
            // int num_records;        // 当前页面中当前已经存储的记录个数，初始化为0 
        char *bitmap;               // 调用init初始化全为0
        char *slots;                // slots指向page里边的内容，清空所有数据，因为page为新创建的，bitmap全0代表无有效数据，数据部分不需要初始化
    */
    file_hdr_.num_pages += 1;
    file_hdr_.first_free_page_no = page->get_page_id().page_no;
    RmPageHandle page_handle(&file_hdr_, page);
    page_handle.page_hdr->next_free_page_no = -1;
    page_handle.page_hdr->num_records = 0;
    Bitmap::init(page_handle.bitmap, file_hdr_.bitmap_size);
    // 这里只写回文件头，写回页面头在使用的查看、插入、删除中调用unpin写回
    disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
    //把页的元数据也写进磁盘，初始化需要使用
    disk_manager_->write_page(fd_, pageid->page_no, page->get_data(), PAGE_SIZE);

     
    return page_handle;
}

/**
 * @brief 创建或获取一个空闲的page handle
 *
 * @return RmPageHandle 返回生成的空闲page handle
 * @note pin the page, remember to unpin it outside!
 */
RmPageHandle RmFileHandle::create_page_handle() {
    // Todo:
    // 1. 判断file_hdr_中是否还有空闲页
    //     1.1 没有空闲页：使用缓冲池来创建一个新page；可直接调用create_new_page_handle()
    //     1.2 有空闲页：直接获取第一个空闲页
    // 2. 生成page handle并返回给上层
    if(file_hdr_.first_free_page_no == INVALID_PAGE_ID)
    {   
        // 没有空闲页
        return create_new_page_handle();
    }
    else
    {   
        RmPageHandle page_handle = fetch_page_handle(file_hdr_.first_free_page_no);
        return page_handle;
    }
}

/**
 * @description: 当一个页面从没有空闲空间的状态变为有空闲空间状态时，更新文件头和页头中空闲页面相关的元数据
 */
void RmFileHandle::release_page_handle(RmPageHandle&page_handle) {
    // Todo:
    // 当page从已满变成未满，考虑如何更新：
    // 1. page_handle.page_hdr->next_free_page_no
    // 2. file_hdr_.first_free_page_no
    
    // 0.1如果所有页本来都满
    if(file_hdr_.first_free_page_no == INVALID_PAGE_ID)
    {
        file_hdr_.first_free_page_no = page_handle.page->get_page_id().page_no;
        disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
    }
    // 0.2如果first_free_page_no指向的位置有空闲记录空间，删了其他一页，将first->next->next...一直到next为-1的页，将这个页next指向刚删除完记录的这页
    else
    {   
        for(int i = file_hdr_.first_free_page_no; /*符合条件直接break*/ ; /*下一个的条件在else里边*/)
        {
            RmPageHandle iter = fetch_page_handle(i);
            if(iter.page_hdr->next_free_page_no == INVALID_PAGE_ID)
            {
                iter.page_hdr->next_free_page_no = page_handle.page->get_page_id().page_no;
                break;
            }
            else
            {
                i = iter.page_hdr->next_free_page_no;
            }
        }
    }
}

void RmFileHandle::update_page_lsn(int page_no, lsn_t lsn) const{
    // 1.如果page_no invalid，抛出PageNotExistError异常
    if(page_no >= RmFileHandle::file_hdr_.num_pages){
        throw PageNotExistError("",page_no);
    }
    // 2.通过buffer pool manager获取指定pageId的page
    PageId page_id;
    page_id.page_no = page_no;
    page_id.fd = fd_;
    Page *page = buffer_pool_manager_->fetch_page(page_id);

    if(page == nullptr){
        throw PageNotExistError("",page_no);
    }

    // 3. 更新lsn
    page->set_page_lsn(lsn);

    // 4. unpin该page
    buffer_pool_manager_->unpin_page(page_id,true);
}

void RmFileHandle::load_record(char* buf, int record_num) {
    RmPageHandle page_handle = create_page_handle();
    char* data = page_handle.slots;
    int rec_num = record_num == 0? file_hdr_.num_records_per_page:record_num;
    std::memcpy(data, buf, rec_num* file_hdr_.record_size);

    for (int i = 0; i < rec_num; i++) {
        Bitmap::set(page_handle.bitmap, i);
    }

    // Update page header
    assert(page_handle.page_hdr->num_records == 0);
    page_handle.page_hdr->num_records = rec_num;

    // Update file header
    if (page_handle.page_hdr->num_records == file_hdr_.num_records_per_page) {
        file_hdr_.first_free_page_no = INVALID_PAGE_ID;
    }
    else{
        file_hdr_.first_free_page_no = page_handle.page->get_page_id().page_no;
    }

    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
}

void RmFileHandle::write_file_hdr() {
    disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char*)&file_hdr_, sizeof(file_hdr_));
}

int RmFileHandle::get_records_num() {
    int records_num = 0;
    for(int i = RM_FIRST_RECORD_PAGE; i < file_hdr_.num_pages; i++)
    {
        RmPageHandle page_handle = fetch_page_handle(i);
        records_num += page_handle.page_hdr->num_records; 
    }
    return records_num;
}